from boa3.boa3 import Boa3
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.testengine import TestEngine


class TestUnion(BoaTest):
    default_folder: str = 'test_sc/union_test'

    def test_union_return(self):
        integer = Integer(42).to_byte_array()
        string = String('42').to_bytes()
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.JMPIFNOT
            + Integer(5).to_byte_array(signed=True, min_length=1)
            + Opcode.PUSHINT8 + integer  # return 42
            + Opcode.RET
            + Opcode.PUSHDATA1  # a = b'\x01\x02\x03'
            + Integer(len(string)).to_byte_array(min_length=1)
            + string
            + Opcode.RET        # return
        )

        path = self.get_contract_path('UnionReturn.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', True)
        self.assertEqual(42, result)

        result = self.run_smart_contract(engine, path, 'main', False)
        self.assertEqual('42', result)

    def test_union_variable_reassign(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x03'
            + b'\x00'
            + Opcode.PUSH2      # a = 2
            + Opcode.STLOC0
            + Opcode.PUSH2      # b = a
            + Opcode.STLOC1
            + Opcode.PUSH2      # c = [a, b]
            + Opcode.PUSH2
            + Opcode.PUSH2
            + Opcode.PACK
            + Opcode.STLOC2
            + Opcode.LDLOC2     # b = c
            + Opcode.STLOC1
            + Opcode.RET        # return
        )

        path = self.get_contract_path('UnionVariableReassign.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_union_variable_argument(self):
        path = self.get_contract_path('UnionVariableArgument.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', 'unittest')
        self.assertEqual('string', result)

        result = self.run_smart_contract(engine, path, 'main', False)
        self.assertEqual('boolean', result)

    def test_union_isinstance_validation(self):
        path = self.get_contract_path('UnionIsInstanceValidation.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', 'unittest')
        self.assertEqual('unittest', result)

        result = self.run_smart_contract(engine, path, 'main', False)
        self.assertEqual('boolean', result)

    def test_union_int_none(self):
        path = self.get_contract_path('UnionIntNone.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(42, result)



import ast
from typing import Dict, List, Tuple

from boa3.model.builtin.method.listmethod import ListMethod
from boa3.model.type.itype import IType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class ListSequenceMethod(ListMethod):

    def __init__(self, sequence_type: IType = None):
        from boa3.model.type.type import Type
        if sequence_type is None:
            sequence_type = Type.sequence

        args: Dict[str, Variable] = {
            'value': Variable(sequence_type),
        }

        value_default = ast.parse("{0}".format(Type.sequence.default_value)
                                  ).body[0].value

        return_type = Type.list.build_collection(sequence_type.value_type)

        super().__init__(args, return_type, [value_default])

    @property
    def prepare_for_packing(self) -> List[Tuple[Opcode, bytes]]:

        if self._prepare_for_packing is None:

            self._prepare_for_packing = [(Opcode.UNPACK, b'')]

        return super().prepare_for_packing



from boa3.exception import CompilerError
from boa3.neo3.contracts import FindOptions
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.testengine import TestEngine


class TestPythonOperation(BoaTest):
    default_folder: str = 'test_sc/python_operation_test'

    # region Membership

    def test_in_bytes(self):
        path = self.get_contract_path('BytesIn.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', b'123', b'1234')
        self.assertEqual(b'123' in b'1234', result)

        result = self.run_smart_contract(engine, path, 'main', b'42', b'1234')
        self.assertEqual(b'42' in b'1234', result)

        result = self.run_smart_contract(engine, path, 'main', b'34', b'1234')
        self.assertEqual(b'34' in b'1234', result)

    def test_bytes_membership_mismatched_type(self):
        path = self.get_contract_path('BytesMembershipMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_int_in_bytes(self):
        path = self.get_contract_path('BytesMembershipWithInt.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', 1, b'1234')
        self.assertEqual(1 in b'1234', result)

        result = self.run_smart_contract(engine, path, 'main', 50, b'1234')
        self.assertEqual(50 in b'1234', result)

    def test_in_dict(self):
        path = self.get_contract_path('DictIn.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', 1, {1: '2', '4': 8})
        self.assertEqual(1 in {1: '2', '4': 8}, result)

        result = self.run_smart_contract(engine, path, 'main', '1', {1: '2', '4': 8})
        self.assertEqual('1' in {1: '2', '4': 8}, result)

        result = self.run_smart_contract(engine, path, 'main', 8, {1: '2', '4': 8})
        self.assertEqual(8 in {1: '2', '4': 8}, result)

        result = self.run_smart_contract(engine, path, 'main', '4', {1: '2', '4': 8})
        self.assertEqual('4' in {1: '2', '4': 8}, result)

    def test_dict_membership_mismatched_type(self):
        path = self.get_contract_path('DictMembershipMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_in_list(self):
        path = self.get_contract_path('ListIn.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', 1, [1, 2, '3', '4'])
        self.assertEqual(1 in [1, 2, '3', '4'], result)

        result = self.run_smart_contract(engine, path, 'main', 3, [1, 2, '3', '4'])
        self.assertEqual(3 in [1, 2, '3', '4'], result)

        result = self.run_smart_contract(engine, path, 'main', '4', [1, 2, '3', '4'])
        self.assertEqual('4' in [1, 2, '3', '4'], result)

    def test_list_membership_mismatched_type(self):
        path = self.get_contract_path('ListMembershipMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_in_str(self):
        path = self.get_contract_path('StringIn.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', '123', '1234')
        self.assertEqual('123' in '1234', result)

        result = self.run_smart_contract(engine, path, 'main', '42', '1234')
        self.assertEqual('42' in '1234', result)

    def test_str_membership_mismatched_type(self):
        path = self.get_contract_path('StringMembershipMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_in_tuple(self):
        path = self.get_contract_path('TupleIn.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', 1, (1, 2, '3', '4'))
        self.assertEqual(1 in (1, 2, '3', '4'), result)

        result = self.run_smart_contract(engine, path, 'main', 3, (1, 2, '3', '4'))
        self.assertEqual(3 in (1, 2, '3', '4'), result)

        result = self.run_smart_contract(engine, path, 'main', '4', (1, 2, '3', '4'))
        self.assertEqual('4' in (1, 2, '3', '4'), result)

    def test_tuple_membership_mismatched_type(self):
        path = self.get_contract_path('TupleMembershipMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_in_typed_dict_builtin_type(self):
        path = self.get_contract_path('TypedDictBuiltinTypeIn.py')
        engine = TestEngine()

        element = FindOptions.VALUES_ONLY
        dict_ = {FindOptions.NONE: 'FindOptions.NONE', FindOptions.DESERIALIZE_VALUES: 'FindOptions.DESERIALIZE_VALUES'}
        result = self.run_smart_contract(engine, path, 'main', element, dict_)
        self.assertEqual(element in dict_, result)

        element = FindOptions.PICK_FIELD_0
        dict_ = {FindOptions.NONE: 'FindOptions.NONE', FindOptions.DESERIALIZE_VALUES: 'FindOptions.DESERIALIZE_VALUES'}
        result = self.run_smart_contract(engine, path, 'main', element, dict_)
        self.assertEqual(element in dict_, result)

    def test_in_typed_dict(self):
        path = self.get_contract_path('TypedDictIn.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', 1, {1: '2', 4: '8'})
        self.assertEqual(1 in {1: '2', 4: '8'}, result)

        result = self.run_smart_contract(engine, path, 'main', 3, {1: '2', 4: '8'})
        self.assertEqual(3 in {1: '2', 4: '8'}, result)

    def test_in_typed_list_builtin_type(self):
        path = self.get_contract_path('TypedListBuiltinTypeIn.py')
        engine = TestEngine()

        element = FindOptions.VALUES_ONLY
        list_ = [FindOptions.NONE, FindOptions.VALUES_ONLY, FindOptions.DESERIALIZE_VALUES, FindOptions.KEYS_ONLY]
        result = self.run_smart_contract(engine, path, 'main', element, list_)
        self.assertEqual(element in list_, result)

        element = FindOptions.PICK_FIELD_0
        list_ = [FindOptions.NONE, FindOptions.VALUES_ONLY, FindOptions.DESERIALIZE_VALUES, FindOptions.KEYS_ONLY]
        result = self.run_smart_contract(engine, path, 'main', element, list_)
        self.assertEqual(element in list_, result)

    def test_in_typed_list(self):
        path = self.get_contract_path('TypedListIn.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', 1, [1, 2, 3, 4])
        self.assertEqual(1 in [1, 2, 3, 4], result)

        result = self.run_smart_contract(engine, path, 'main', 6, [1, 2, 3, 4])
        self.assertEqual(6 in [1, 2, 3, 4], result)

    def test_in_typed_tuple_builtin_type(self):
        path = self.get_contract_path('TypedTupleBuiltinTypeIn.py')
        engine = TestEngine()

        element = FindOptions.VALUES_ONLY
        tuple_ = (FindOptions.NONE, FindOptions.VALUES_ONLY, FindOptions.DESERIALIZE_VALUES, FindOptions.KEYS_ONLY)
        result = self.run_smart_contract(engine, path, 'main', element, tuple_)
        self.assertEqual(element in tuple_, result)

        element = FindOptions.PICK_FIELD_0
        tuple_ = (FindOptions.NONE, FindOptions.VALUES_ONLY, FindOptions.DESERIALIZE_VALUES, FindOptions.KEYS_ONLY)
        result = self.run_smart_contract(engine, path, 'main', element, tuple_)
        self.assertEqual(element in tuple_, result)

    def test_in_typed_tuple(self):
        path = self.get_contract_path('TypedTupleIn.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', 1, (1, 2, 3, 4))
        self.assertEqual(1 in (1, 2, 3, 4), result)

        result = self.run_smart_contract(engine, path, 'main', 6, (1, 2, 3, 4))
        self.assertEqual(6 in (1, 2, 3, 4), result)

    # endregion

    # region NotMembership

    def test_not_in_bytes(self):
        path = self.get_contract_path('BytesNotIn.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', b'123', b'1234')
        self.assertEqual(b'123' not in b'1234', result)

        result = self.run_smart_contract(engine, path, 'main', b'42', b'1234')
        self.assertEqual(b'42' not in b'1234', result)

    def test_not_in_dict(self):
        path = self.get_contract_path('DictNotIn.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', 1, {1: '2', '4': 8})
        self.assertEqual(1 not in {1: '2', '4': 8}, result)

        result = self.run_smart_contract(engine, path, 'main', '1', {1: '2', '4': 8})
        self.assertEqual('1' not in {1: '2', '4': 8}, result)

        result = self.run_smart_contract(engine, path, 'main', 8, {1: '2', '4': 8})
        self.assertEqual(8 not in {1: '2', '4': 8}, result)

        result = self.run_smart_contract(engine, path, 'main', '4', {1: '2', '4': 8})
        self.assertEqual('4' not in {1: '2', '4': 8}, result)

    def test_not_in_list(self):
        path = self.get_contract_path('ListNotIn.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', 1, [1, 2, '3', '4'])
        self.assertEqual(1 not in [1, 2, '3', '4'], result)

        result = self.run_smart_contract(engine, path, 'main', 3, [1, 2, '3', '4'])
        self.assertEqual(3 not in [1, 2, '3', '4'], result)

        result = self.run_smart_contract(engine, path, 'main', '4', [1, 2, '3', '4'])
        self.assertEqual('4' not in [1, 2, '3', '4'], result)

    def test_not_in_str(self):
        path = self.get_contract_path('StringNotIn.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', '123', '1234')
        self.assertEqual('123' not in '1234', result)

        result = self.run_smart_contract(engine, path, 'main', '42', '1234')
        self.assertEqual('42' not in '1234', result)

    def test_not_in_tuple(self):
        path = self.get_contract_path('TupleNotIn.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', 1, (1, 2, '3', '4'))
        self.assertEqual(1 not in (1, 2, '3', '4'), result)

        result = self.run_smart_contract(engine, path, 'main', 3, (1, 2, '3', '4'))
        self.assertEqual(3 not in (1, 2, '3', '4'), result)

        result = self.run_smart_contract(engine, path, 'main', '4', (1, 2, '3', '4'))
        self.assertEqual('4' not in (1, 2, '3', '4'), result)

    def test_not_in_typed_dict_builtin_type(self):
        path = self.get_contract_path('TypedDictBuiltinTypeNotIn.py')
        engine = TestEngine()

        element = FindOptions.VALUES_ONLY
        dict_ = {FindOptions.NONE: 'FindOptions.NONE', FindOptions.DESERIALIZE_VALUES: 'FindOptions.DESERIALIZE_VALUES'}
        result = self.run_smart_contract(engine, path, 'main', element, dict_)
        self.assertEqual(element not in dict_, result)

        element = FindOptions.PICK_FIELD_0
        dict_ = {FindOptions.NONE: 'FindOptions.NONE', FindOptions.DESERIALIZE_VALUES: 'FindOptions.DESERIALIZE_VALUES'}
        result = self.run_smart_contract(engine, path, 'main', element, dict_)
        self.assertEqual(element not in dict_, result)

    def test_not_in_typed_dict(self):
        path = self.get_contract_path('TypedDictNotIn.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', 1, {1: '2', 4: '8'})
        self.assertEqual(1 not in {1: '2', 4: '8'}, result)

        result = self.run_smart_contract(engine, path, 'main', 3, {1: '2', 4: '8'})
        self.assertEqual(3 not in {1: '2', 4: '8'}, result)

    def test_not_in_typed_list_builtin_type(self):
        path = self.get_contract_path('TypedListBuiltinTypeNotIn.py')
        engine = TestEngine()

        element = FindOptions.VALUES_ONLY
        list_ = [FindOptions.NONE, FindOptions.VALUES_ONLY, FindOptions.DESERIALIZE_VALUES, FindOptions.KEYS_ONLY]
        result = self.run_smart_contract(engine, path, 'main', element, list_)
        self.assertEqual(element not in list_, result)

        element = FindOptions.PICK_FIELD_0
        list_ = [FindOptions.NONE, FindOptions.VALUES_ONLY, FindOptions.DESERIALIZE_VALUES, FindOptions.KEYS_ONLY]
        result = self.run_smart_contract(engine, path, 'main', element, list_)
        self.assertEqual(element not in list_, result)

    def test_not_in_typed_list(self):
        path = self.get_contract_path('TypedListNotIn.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', 1, [1, 2, 3, 4])
        self.assertEqual(1 not in [1, 2, 3, 4], result)

        result = self.run_smart_contract(engine, path, 'main', 6, [1, 2, 3, 4])
        self.assertEqual(6 not in [1, 2, 3, 4], result)

    def test_not_in_typed_tuple_builtin_type(self):
        path = self.get_contract_path('TypedTupleBuiltinTypeNotIn.py')
        engine = TestEngine()

        element = FindOptions.VALUES_ONLY
        tuple_ = (FindOptions.NONE, FindOptions.VALUES_ONLY, FindOptions.DESERIALIZE_VALUES, FindOptions.KEYS_ONLY)
        result = self.run_smart_contract(engine, path, 'main', element, tuple_)
        self.assertEqual(element not in tuple_, result)

        element = FindOptions.PICK_FIELD_0
        tuple_ = (FindOptions.NONE, FindOptions.VALUES_ONLY, FindOptions.DESERIALIZE_VALUES, FindOptions.KEYS_ONLY)
        result = self.run_smart_contract(engine, path, 'main', element, tuple_)
        self.assertEqual(element not in tuple_, result)

    def test_not_in_typed_tuple(self):
        path = self.get_contract_path('TypedTupleNotIn.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', 1, (1, 2, 3, 4))
        self.assertEqual(1 not in (1, 2, 3, 4), result)

        result = self.run_smart_contract(engine, path, 'main', 6, (1, 2, 3, 4))
        self.assertEqual(6 not in (1, 2, 3, 4), result)

    # endregion



from typing import List, Tuple

from boa3.model.operation.binary.binaryoperation import BinaryOperation
from boa3.model.operation.operator import Operator
from boa3.model.type.type import IType, Type
from boa3.neo.vm.opcode import OpcodeHelper
from boa3.neo.vm.opcode.Opcode import Opcode


class Modulo(BinaryOperation):
    """
    A class used to represent a numeric modulo operation

    :ivar operator: the operator of the operation. Inherited from :class:`IOperation`
    :ivar left: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar right: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar result: the result type of the operation.  Inherited from :class:`IOperation`
    """
    _valid_types: List[IType] = [Type.int]

    def __init__(self, left: IType = Type.int, right: IType = None):
        self.operator: Operator = Operator.Mod
        super().__init__(left, right)

    def validate_type(self, *types: IType) -> bool:
        if len(types) != self.number_of_operands:
            return False
        left: IType = types[0]
        right: IType = types[1]

        return left == right and any(_type.is_type_of(left) for _type in self._valid_types)

    def _get_result(self, left: IType, right: IType) -> IType:
        if self.validate_type(left, right):
            return left
        else:
            return Type.none

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        jmp_place_holder = (Opcode.JMP, b'')

        mod = [
            (Opcode.SWAP, b''),  # neo's mod has a different result from python's mod in some cases
            (Opcode.OVER, b''),
            (Opcode.MOD, b''),
            (Opcode.DUP, b''),
            (Opcode.SIGN, b''),
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.SIGN, b''),
            (Opcode.OVER, b''),      # if the result is not zero and the sign is different than the second operator
            (Opcode.NUMEQUAL, b''),  # the result is different
            (Opcode.SWAP, b''),
            (Opcode.PUSH0, b''),
            (Opcode.NUMEQUAL, b''),  # the result is different
            (Opcode.BOOLOR, b''),
            jmp_place_holder,
        ]

        from boa3.compiler.codegenerator import get_bytes_count
        if_negative = [
            (Opcode.ADD, b''),
            jmp_place_holder
        ]
        num_jmp_code = get_bytes_count(if_negative)
        mod[-1] = OpcodeHelper.get_jump_and_data(Opcode.JMPIF, num_jmp_code, True)

        else_negative = [
            (Opcode.NIP, b'')
        ]
        num_jmp_code = get_bytes_count(else_negative)
        if_negative[-1] = OpcodeHelper.get_jump_and_data(Opcode.JMP, num_jmp_code, True)

        return mod + if_negative + else_negative



from boa3.builtin.interop.role.roletype import Role
from boa3.builtin.type import ECPoint, UInt160


class RoleManagement:
    """
    A class used to represent the RoleManagement native contract
    """

    hash: UInt160

    @classmethod
    def get_designated_by_role(cls, role: Role, index: int) -> ECPoint:
        """
        Gets the list of nodes for the specified role.

        :param role: the type of the role
        :type role: Role
        :param index: the index of the block to be queried
        :type index: int

        :return: the public keys of the nodes
        :rtype: ECPoint
        """
        pass



import unittest

from boa3.neo.vm.type.AbiType import AbiType
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest
from test_runner import utils
from test_runner.neo_test_runner import NeoTestRunner


class TestTestRunner(BoaTest):

    @unittest.skip("don't run while circleci config is not updated with test runner")
    def test_run(self):
        path = self.get_contract_path('test_sc/generation_test', 'GenerationWithDecorator.py')
        self.compile_and_save(path)
        path = path.replace('.py', '.nef')

        runner = NeoTestRunner()
        result = runner.run(path, 'Sub', 50, 20)
        self.assertEqual(30, result)

    def test_int_value_to_invoke_parameter(self):
        expected_result = 5
        result = utils.value_to_parameter(5)
        self.assertEqual(expected_result, result)

        expected_result = 5000000
        result = utils.value_to_parameter(5000000)
        self.assertEqual(expected_result, result)

    def test_bool_value_to_invoke_parameter(self):
        expected_result = True
        result = utils.value_to_parameter(True)
        self.assertEqual(expected_result, result)

        expected_result = False
        result = utils.value_to_parameter(False)
        self.assertEqual(expected_result, result)

    def test_str_value_to_invoke_parameter(self):
        expected_result = 'unittest'
        result = utils.value_to_parameter('unittest')
        self.assertEqual(expected_result, result)

        long_string = ('Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nullam accumsan magna eu massa '
                       'vulputate bibendum. Aliquam commodo euismod tristique. Sed purus erat, pretium ut interdum '
                       'et, aliquet sed mauris. Curabitur vitae turpis euismod, hendrerit mi a, rhoncus justo. Mauris '
                       'sollicitudin, nisl sit amet feugiat pharetra, odio ligula congue tellus, vel pellentesque '
                       'libero leo id dui. Morbi vel risus vehicula, consectetur mauris eget, gravida ligula. '
                       'Maecenas aliquam velit sit amet nisi ultricies, ac sollicitudin nisi mollis. Lorem ipsum '
                       'dolor sit amet, consectetur adipiscing elit. Ut tincidunt, nisi in ullamcorper ornare, '
                       'est enim dictum massa, id aliquet justo magna in purus.')
        expected_result = long_string
        result = utils.value_to_parameter(long_string)
        self.assertEqual(expected_result, result)

    def test_none_value_to_invoke_parameter(self):
        expected_result = None
        result = utils.value_to_parameter(None)
        self.assertEqual(expected_result, result)

    def test_bytes_value_to_invoke_parameter(self):
        expected_result = '0x0102030405'
        result = utils.value_to_parameter(b'\x01\x02\x03\x04\x05')
        self.assertEqual(expected_result, result)

        long_bytes = (String('Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nullam accumsan magna eu massa '
                             'vulputate bibendum. Aliquam commodo euismod tristique. Sed purus erat, pretium ut '
                             'interdum et, aliquet sed mauris. Curabitur vitae turpis euismod, hendrerit mi a, '
                             'rhoncus justo. Mauris sollicitudin, nisl sit amet feugiat pharetra, odio ligula congue '
                             'tellus, vel pellentesque libero leo id dui. Morbi vel risus vehicula, consectetur '
                             'mauris eget, gravida ligula. Maecenas aliquam velit sit amet nisi ultricies, '
                             'ac sollicitudin nisi mollis. Lorem ipsum dolor sit amet, consectetur adipiscing elit. '
                             'Ut tincidunt, nisi in ullamcorper ornare, est enim dictum massa, id aliquet justo magna '
                             'in purus.')
                      .to_bytes())

        expected_result = utils.bytes_to_hex(long_bytes)
        result = utils.value_to_parameter(long_bytes)
        self.assertEqual(expected_result, result)

    def test_list_value_to_invoke_parameter(self):
        expected_result = [1, 2, 3]
        result = utils.value_to_parameter([1, 2, 3])
        self.assertEqual(expected_result, result)

        expected_result = [True, 2, '3']
        result = utils.value_to_parameter([True, 2, '3'])
        self.assertEqual(expected_result, result)

    def test_tuple_value_to_invoke_parameter(self):
        expected_result = ['0x0102', '0x0304', '0x0506']
        result = utils.value_to_parameter((b'\x01\x02', b'\x03\x04', b'\x05\x06'))
        self.assertEqual(expected_result, result)

        expected_result = [True, 2, '3']
        result = utils.value_to_parameter((True, 2, '3'))
        self.assertEqual(expected_result, result)

    def test_dict_value_to_invoke_parameter(self):
        expected_result = {
            'type': AbiType.Map.value,
            'value': [
                {'key': '0x61', 'value': 1},
                {'key': '0x62', 'value': False}
            ]
        }
        result = utils.value_to_parameter({
            b'a': 1,
            b'b': False
        })
        self.assertEqual(expected_result, result)



from enum import Enum
from typing import Dict, List

from boa3.model.builtin.interop.blockchain import *
from boa3.model.builtin.interop.contract import *
from boa3.model.builtin.interop.contract.contractmanifest import *
from boa3.model.builtin.interop.crypto import *
from boa3.model.builtin.interop.iterator import *
from boa3.model.builtin.interop.json import *
from boa3.model.builtin.interop.nativecontract import *
from boa3.model.builtin.interop.oracle import *
from boa3.model.builtin.interop.policy import *
from boa3.model.builtin.interop.role import *
from boa3.model.builtin.interop.runtime import *
from boa3.model.builtin.interop.stdlib import *
from boa3.model.builtin.interop.storage import *
from boa3.model.event import Event
from boa3.model.identifiedsymbol import IdentifiedSymbol
from boa3.model.imports.package import Package

__all__ = ['Interop',
           'InteropPackage'
           ]


class InteropPackage(str, Enum):
    Blockchain = 'blockchain'
    Contract = 'contract'
    Crypto = 'crypto'
    Iterator = 'iterator'
    Json = 'json'
    Oracle = 'oracle'
    Policy = 'policy'
    Role = 'role'
    Runtime = 'runtime'
    Stdlib = 'stdlib'
    Storage = 'storage'


class Interop:

    @classmethod
    def interop_symbols(cls, package: str = None) -> List[IdentifiedSymbol]:
        if package in InteropPackage.__members__.values():
            return cls._interop_symbols[package]

        lst: List[IdentifiedSymbol] = []
        for symbols in cls._interop_symbols.values():
            lst.extend(symbols)
        return lst

    @classmethod
    def interop_events(cls) -> List[Event]:
        lst: List[Event] = []
        for symbols in cls._interop_symbols.values():
            lst.extend([event for event in symbols if isinstance(event, Event)])

        return lst

    # region Interops

    # Interop Types
    BlockType = BlockType.build()
    CallFlagsType = CallFlagsType()
    ContractManifestType = ContractManifestType.build()
    ContractType = ContractType.build()
    FindOptionsType = FindOptionsType()
    Iterator = IteratorType.build()
    NamedCurveType = NamedCurveType()
    NotificationType = NotificationType.build()
    OracleResponseCode = OracleResponseCodeType.build()
    OracleType = OracleClass.build()
    RoleType = RoleType.build()
    SignerType = SignerType.build()
    StorageContextType = StorageContextType.build()
    StorageMapType = StorageMapType.build()
    TransactionType = TransactionType.build()
    TriggerType = TriggerType()
    VMStateType = VMStateType.build()
    WitnessCondition = WitnessConditionType.build()
    WitnessConditionType = WitnessConditionEnumType.build()
    WitnessRuleAction = WitnessRuleActionType.build()
    WitnessRule = WitnessRuleType.build()
    WitnessScope = WitnessScopeType.build()

    # Blockchain Interops
    CurrentHash = CurrentHashProperty()
    CurrentIndex = CurrentIndexProperty()
    GetContract = GetContractMethod(ContractType)
    GetBlock = GetBlockMethod(BlockType)
    GetTransaction = GetTransactionMethod(TransactionType)
    GetTransactionFromBlock = GetTransactionFromBlockMethod(TransactionType)
    GetTransactionHeight = GetTransactionHeightMethod()
    GetTransactionSigners = GetTransactionSignersMethod(SignerType)
    GetTransactionVMState = GetTransactionVMStateMethod(VMStateType)

    # Contract Interops
    CallContract = CallMethod()
    CreateContract = CreateMethod(ContractType)
    CreateMultisigAccount = CreateMultisigAccountMethod()
    CreateStandardAccount = CreateStandardAccountMethod()
    DestroyContract = DestroyMethod()
    GetCallFlags = GetCallFlagsMethod(CallFlagsType)
    GetMinimumDeploymentFee = GetMinimumDeploymentFeeMethod()
    UpdateContract = UpdateMethod()

    # Native Contracts
    GasScriptHash = GasToken
    NeoScriptHash = NeoToken
    ContractManagementScriptHash = ContractManagement
    CryptoLibScriptHash = CryptoLibContract
    LedgerScriptHash = LedgerContract
    OracleScriptHash = OracleContract
    StdLibScriptHash = StdLibContract

    # Crypto Interops
    CheckMultisig = CheckMultisigMethod()
    CheckSig = CheckSigMethod()
    Hash160 = Hash160Method()
    Hash256 = Hash256Method()
    Murmur32 = Murmur32Method()
    Ripemd160 = Ripemd160Method()
    Sha256 = Sha256Method()
    VerifyWithECDsa = VerifyWithECDsaMethod()

    # Iterator Interops
    IteratorCreate = IteratorMethod(Iterator)

    # Json Interops
    JsonDeserialize = JsonDeserializeMethod()
    JsonSerialize = JsonSerializeMethod()

    # Policy Interops
    GetExecFeeFactor = GetExecFeeFactorMethod()
    GetFeePerByte = GetFeePerByteMethod()
    GetStoragePrice = GetStoragePriceMethod()
    IsBlocked = IsBlockedMethod()

    # Role Interops
    GetDesignatedByRole = GetDesignatedByRoleMethod()

    # Runtime Interops
    AddressVersion = AddressVersionProperty()
    BlockTime = BlockTimeProperty()
    BurnGas = BurnGasMethod()
    CallingScriptHash = CallingScriptHashProperty()
    CheckWitness = CheckWitnessMethod()
    EntryScriptHash = EntryScriptHashProperty()
    ExecutingScriptHash = ExecutingScriptHashProperty()
    GasLeft = GasLeftProperty()
    GetNetwork = GetNetworkMethod()
    GetNotifications = GetNotificationsMethod(NotificationType)
    GetRandom = GetRandomMethod()
    GetTrigger = GetTriggerMethod(TriggerType)
    InvocationCounter = InvocationCounterProperty()
    LoadScript = LoadScriptMethod()
    Log = LogMethod()
    Notify = NotifyMethod()
    Platform = PlatformProperty()
    ScriptContainer = ScriptContainerProperty()

    # Stdlib Interops
    Atoi = AtoiMethod()
    Base58CheckDecode = Base58CheckDecodeMethod()
    Base58CheckEncode = Base58CheckEncodeMethod()
    Base58Encode = Base58EncodeMethod()
    Base58Decode = Base58DecodeMethod()
    Base64Encode = Base64EncodeMethod()
    Base64Decode = Base64DecodeMethod()
    Deserialize = DeserializeMethod()
    Itoa = ItoaMethod()
    MemoryCompare = MemoryCompareMethod()
    MemorySearch = MemorySearchMethod()
    Serialize = SerializeMethod()

    # Storage Interops
    StorageDelete = StorageDeleteMethod()
    StorageFind = StorageFindMethod(FindOptionsType)
    StorageGetContext = StorageGetContextMethod(StorageContextType)
    StorageGetReadOnlyContext = StorageGetReadOnlyContextMethod(StorageContextType)
    StorageGet = StorageGetMethod()
    StoragePut = StoragePutMethod()

    # endregion

    # region Packages

    BlockModule = Package(identifier=BlockType.identifier.lower(),
                          types=[BlockType]
                          )

    SignerModule = Package(identifier=SignerType.identifier.lower(),
                           types=[SignerType,
                                  WitnessConditionType,
                                  WitnessCondition,
                                  WitnessRuleAction,
                                  WitnessRule,
                                  WitnessScope
                                  ]
                           )

    TransactionModule = Package(identifier=TransactionType.identifier.lower(),
                                types=[TransactionType]
                                )

    VMStateModule = Package(identifier=VMStateType.identifier.lower(),
                            types=[VMStateType]
                            )

    BlockchainPackage = Package(identifier=InteropPackage.Blockchain,
                                types=[BlockType,
                                       SignerType,
                                       TransactionType,
                                       VMStateType
                                       ],
                                methods=[CurrentHash,
                                         CurrentIndex,
                                         GetBlock,
                                         GetContract,
                                         GetTransaction,
                                         GetTransactionFromBlock,
                                         GetTransactionHeight,
                                         GetTransactionSigners,
                                         GetTransactionVMState
                                         ],
                                packages=[BlockModule,
                                          SignerModule,
                                          TransactionModule,
                                          VMStateModule
                                          ]
                                )

    CallFlagsTypeModule = Package(identifier=f'{CallFlagsType.identifier.lower()}type',
                                  types=[CallFlagsType]
                                  )

    ContractModule = Package(identifier=ContractType.identifier.lower(),
                             types=[ContractType]
                             )

    ContractManifestModule = Package(identifier=ContractManifestType.identifier.lower(),
                                     types=[ContractAbiType.build(),
                                            ContractEventDescriptorType.build(),
                                            ContractGroupType.build(),
                                            ContractManifestType,
                                            ContractMethodDescriptorType.build(),
                                            ContractParameterDefinitionType.build(),
                                            ContractParameterType.build(),
                                            ContractPermissionDescriptorType.build(),
                                            ContractPermissionType.build()
                                            ]
                                     )

    ContractPackage = Package(identifier=InteropPackage.Contract,
                              types=[CallFlagsType,
                                     ContractManifestType,
                                     ContractType
                                     ],
                              properties=[GasScriptHash,
                                          NeoScriptHash
                                          ],
                              methods=[CallContract,
                                       CreateContract,
                                       CreateMultisigAccount,
                                       CreateStandardAccount,
                                       DestroyContract,
                                       GetCallFlags,
                                       GetMinimumDeploymentFee,
                                       UpdateContract
                                       ],
                              packages=[CallFlagsTypeModule,
                                        ContractManifestModule,
                                        ContractModule
                                        ]
                              )

    CryptoPackage = Package(identifier=InteropPackage.Crypto,
                            types=[NamedCurveType],
                            methods=[CheckMultisig,
                                     CheckSig,
                                     Hash160,
                                     Hash256,
                                     Murmur32,
                                     Ripemd160,
                                     Sha256,
                                     VerifyWithECDsa,
                                     ]
                            )

    IteratorPackage = Package(identifier=InteropPackage.Iterator,
                              types=[Iterator],
                              )

    JsonPackage = Package(identifier=InteropPackage.Json,
                          methods=[JsonDeserialize,
                                   JsonSerialize
                                   ]
                          )

    NotificationModule = Package(identifier=NotificationType.identifier.lower(),
                                 types=[NotificationType]
                                 )

    OracleResponseCodeModule = Package(identifier=OracleResponseCode.identifier.lower(),
                                       types=[OracleResponseCode]
                                       )

    OracleModule = Package(identifier=OracleType.identifier.lower(),
                           types=[OracleType]
                           )

    OraclePackage = Package(identifier=InteropPackage.Oracle,
                            types=[OracleResponseCode,
                                   OracleType
                                   ],
                            packages=[OracleModule,
                                      OracleResponseCodeModule
                                      ]
                            )

    TriggerTypeModule = Package(identifier=TriggerType.identifier.lower(),
                                types=[TriggerType]
                                )

    PolicyPackage = Package(identifier=InteropPackage.Policy,
                            methods=[GetExecFeeFactor,
                                     GetFeePerByte,
                                     GetStoragePrice,
                                     IsBlocked
                                     ]
                            )

    RolePackage = Package(identifier=InteropPackage.Role,
                          types=[RoleType],
                          methods=[GetDesignatedByRole]
                          )

    RuntimePackage = Package(identifier=InteropPackage.Runtime,
                             types=[NotificationType,
                                    TriggerType
                                    ],
                             properties=[AddressVersion,
                                         BlockTime,
                                         CallingScriptHash,
                                         ExecutingScriptHash,
                                         GasLeft,
                                         Platform,
                                         InvocationCounter,
                                         EntryScriptHash,
                                         ScriptContainer
                                         ],
                             methods=[BurnGas,
                                      CheckWitness,
                                      GetNetwork,
                                      GetNotifications,
                                      GetRandom,
                                      GetTrigger,
                                      LoadScript,
                                      Log,
                                      Notify
                                      ],
                             packages=[NotificationModule,
                                       TriggerTypeModule
                                       ]
                             )

    FindOptionsModule = Package(identifier=FindOptionsType.identifier.lower(),
                                types=[FindOptionsType]
                                )

    StdlibPackage = Package(identifier=InteropPackage.Stdlib,
                            methods=[Atoi,
                                     Base58CheckDecode,
                                     Base58CheckEncode,
                                     Base58Encode,
                                     Base58Decode,
                                     Base64Encode,
                                     Base64Decode,
                                     Deserialize,
                                     Itoa,
                                     MemoryCompare,
                                     MemorySearch,
                                     Serialize
                                     ]
                            )

    StorageContextModule = Package(identifier=StorageContextType.identifier.lower(),
                                   types=[StorageContextType]
                                   )

    StorageMapModule = Package(identifier=StorageMapType.identifier.lower(),
                               types=[StorageMapType]
                               )

    StoragePackage = Package(identifier=InteropPackage.Storage,
                             types=[FindOptionsType,
                                    StorageContextType,
                                    StorageMapType
                                    ],
                             methods=[StorageDelete,
                                      StorageFind,
                                      StorageGet,
                                      StorageGetContext,
                                      StorageGetReadOnlyContext,
                                      StoragePut
                                      ],
                             packages=[FindOptionsModule,
                                       StorageContextModule,
                                       StorageMapModule
                                       ]
                             )

    # endregion

    package_symbols: List[IdentifiedSymbol] = [
        OracleType,
        BlockchainPackage,
        ContractPackage,
        CryptoPackage,
        IteratorPackage,
        JsonPackage,
        OraclePackage,
        PolicyPackage,
        RolePackage,
        RuntimePackage,
        StdlibPackage,
        StoragePackage
    ]

    _interop_symbols: Dict[InteropPackage, List[IdentifiedSymbol]] = {
        package.identifier: list(package.symbols.values()) for package in package_symbols if
        isinstance(package, Package)
    }



import ast
from typing import Dict, List, Tuple

from boa3.model.builtin.interop.nativecontract import StdLibMethod
from boa3.model.variable import Variable
from boa3.neo.vm.opcode import OpcodeHelper
from boa3.neo.vm.opcode.Opcode import Opcode


class StrSplitMethod(StdLibMethod):
    def __init__(self):
        from boa3.model.type.type import Type
        identifier = 'split'
        syscall = 'stringSplit'
        args: Dict[str, Variable] = {
            'self': Variable(Type.str),
            'sep': Variable(Type.str),
            'maxsplit': Variable(Type.int)
        }
        # whitespace is the default separator
        separator_default = ast.parse("' '").body[0].value
        # maxsplit the default value is -1
        maxsplit_default = ast.parse("-1").body[0].value.operand
        maxsplit_default.n = -1

        neo_internal_args = {
            'str': Variable(Type.str),
            'separator': Variable(Type.str),
            # TODO: include removeEmptyEntries
        }

        super().__init__(identifier, syscall, args, defaults=[separator_default, maxsplit_default],
                         return_type=Type.list.build_collection(Type.str),
                         internal_call_args=len(neo_internal_args))

    @property
    def generation_order(self) -> List[int]:
        # the original string must be the top value in the stack
        indexes = list(range(len(self.args)))
        str_index = list(self.args).index('self')

        if indexes[-1] != str_index:
            # context must be the last generated argument
            indexes.remove(str_index)
            indexes.append(str_index)
        return indexes

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.compiler.codegenerator import get_bytes_count
        from boa3.model.type.type import Type

        jmp_place_holder = (Opcode.JMP, b'\x01')

        preserver_args_from_array = [   # copies split and maxsplit args on the stack
            OpcodeHelper.get_push_and_data(2),
            (Opcode.PICK, b''),
            (Opcode.SWAP, b''),
        ]

        neo_strsplit_method = super()._opcode

        verify_maxsplit = [     # verifies if there is a maxsplit
            (Opcode.OVER, b''),
            (Opcode.PUSHM1, b''),
            jmp_place_holder    # if maxsplit <= -1 skip concatenation and clean the stack
        ]

        while_verify = [        # verifies if len(array) <= maxsplit + 1, if it is jump out of while
            (Opcode.DUP, b''),
            (Opcode.SIZE, b''),
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.INC, b''),
            jmp_place_holder    # go clean the stack
        ]

        concatenate_array = [
            (Opcode.DUP, b''),
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            (Opcode.OVER, b''),
            (Opcode.POPITEM, b''),
            (Opcode.CAT, b''),
            (Opcode.OVER, b''),
            (Opcode.POPITEM, b''),
            (Opcode.SWAP, b''),
            (Opcode.CAT, b''),
            (Opcode.CONVERT, Type.str.stack_item),
            (Opcode.APPEND, b''),
            # go to while_verify
        ]

        num_jmp_code = -get_bytes_count(while_verify + concatenate_array)
        jmp_back_to_while = OpcodeHelper.get_jump_and_data(Opcode.JMP, num_jmp_code)
        concatenate_array.append(jmp_back_to_while)

        num_jmp_code = get_bytes_count(while_verify + concatenate_array)
        jmp_to_clean_from_verify_maxsplit = OpcodeHelper.get_jump_and_data(Opcode.JMPLE, num_jmp_code, True)
        verify_maxsplit[-1] = jmp_to_clean_from_verify_maxsplit

        num_jmp_code = get_bytes_count(concatenate_array)
        jmp_to_clean_from_while_verify = OpcodeHelper.get_jump_and_data(Opcode.JMPLE, num_jmp_code, True)
        while_verify[-1] = jmp_to_clean_from_while_verify

        clean_stack = [
            (Opcode.REVERSE3, b''),
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),
        ]

        return (
            preserver_args_from_array +
            neo_strsplit_method +
            verify_maxsplit +
            while_verify +
            concatenate_array +
            clean_stack
        )



from boa3.builtin.compile_time import public
from boa3_test.test_sc.import_test.class_import import ImportUserClass


@public
def build_example_object() -> str:
    return ImportUserClass.build_example_object().var_str



from typing import Dict, Optional

from boa3.model.builtin.classmethod.popmethod import PopMethod
from boa3.model.type.itype import IType
from boa3.model.variable import Variable


class PopDictMethod(PopMethod):

    def __init__(self, arg_value: Optional[IType] = None):
        from boa3.model.type.type import Type

        if not Type.dict.is_type_of(arg_value):
            arg_value = Type.dict

        args: Dict[str, Variable] = {
            'self': Variable(arg_value),
            'key': Variable(arg_value.valid_key)
        }

        super().__init__(args, return_type=arg_value.value_type)



from typing import Dict

from boa3.model.builtin.interop.interopmethod import InteropMethod
from boa3.model.variable import Variable


class StorageContextAsReadOnlyMethod(InteropMethod):

    def __init__(self):
        from boa3.model.builtin.interop.storage.storagecontext.storagecontexttype import _StorageContext

        identifier = 'as_read_only'
        syscall = 'System.Storage.AsReadOnly'
        args: Dict[str, Variable] = {'self': Variable(_StorageContext)}
        super().__init__(identifier, syscall, args, return_type=_StorageContext)



import re
from typing import Dict, List

from boa3.analyser import supportedstandard
from boa3.analyser.astanalyser import IAstAnalyser
from boa3.exception import CompilerError
from boa3.model.event import Event
from boa3.model.imports import importsymbol
from boa3.model.method import Method
from boa3.model.symbol import ISymbol


class StandardAnalyser(IAstAnalyser):
    """
    This class is responsible for the checking if the given contract standards are implemented

    :ivar symbols: a dictionary that maps the global symbols.
    """

    def __init__(self, analyser, symbol_table: Dict[str, ISymbol], log: bool = False):
        from boa3.builtin.compile_time import NeoMetadata

        super().__init__(analyser.ast_tree, analyser.filename, log=log)

        self.symbols: Dict[str, ISymbol] = symbol_table

        if isinstance(analyser.metadata, NeoMetadata):
            # filter only strings
            analyser.metadata.supported_standards = [standard for standard in analyser.metadata.supported_standards
                                                     if isinstance(standard, str)]
            standards = analyser.metadata.supported_standards
        else:
            standards = []

        self.standards: List[str] = standards
        self._filter_standards_names()
        self._validate_standards()
        self._check_other_implemented_standards()

    def _filter_standards_names(self):
        """
        Converts all the Neo standards names to upper kebab case
        """
        filtered_standards = set()
        for standard in self.standards:
            standard = standard.strip()
            if standard.upper().startswith('NEP'):
                rebab_case = re.sub(r'([A-Z]+|(.\d+)+|\d+)[^a-zA-Z\d]?', r'\g<1>-', standard.upper())
                if rebab_case.endswith('-'):
                    rebab_case = rebab_case[:-1]
                standard = rebab_case

            filtered_standards.add(standard)

        self.standards.clear()
        self.standards.extend(filtered_standards)

    def get_methods_by_display_name(self, method_id: str) -> List[Method]:
        methods = []
        for symbol_id, symbol in self.symbols.items():
            if isinstance(symbol, Method) and symbol.is_public:
                display_name = symbol.external_name if symbol.external_name is not None else symbol_id
                if display_name == method_id:
                    methods.append(symbol)
        return methods

    def _validate_standards(self):
        for standard in self.standards:
            if standard in supportedstandard.neo_standards:
                current_standard = supportedstandard.neo_standards[standard]

                # validate standard's methods
                for standard_method in current_standard.methods:
                    method_id = standard_method.external_name
                    is_implemented = False

                    found_methods = self.get_methods_by_display_name(method_id)
                    for method in found_methods:
                        if isinstance(method, Method) and current_standard.match_definition(standard_method, method):
                            is_implemented = True
                            break

                    if not is_implemented:
                        self._log_error(
                            CompilerError.MissingStandardDefinition(standard, method_id, standard_method)
                        )

                # validate standard's events
                events = [symbol for symbol in self.symbols.values() if isinstance(symbol, Event)]
                # imported events should be included in the validation
                for imported in self._get_all_imports():
                    events.extend([event for event in imported.all_symbols.values() if isinstance(event, Event)])

                for standard_event in current_standard.events:
                    is_implemented = False
                    for event in events:
                        if (event.name == standard_event.name
                                and current_standard.match_definition(standard_event, event)):
                            is_implemented = True
                            break

                    if not is_implemented:
                        self._log_error(
                            CompilerError.MissingStandardDefinition(standard,
                                                                    standard_event.name,
                                                                    standard_event)
                        )

                # validate optional methods
                for optional_method in current_standard.optionals:
                    method_id = optional_method.external_name
                    is_implemented = False

                    found_methods = self.get_methods_by_display_name(method_id)
                    for method in found_methods:
                        if isinstance(method, Method) and current_standard.match_definition(optional_method, method):
                            is_implemented = True
                            break

                    if found_methods and not is_implemented:
                        self._log_error(
                            CompilerError.MissingStandardDefinition(standard, method_id, optional_method)
                        )

    def _get_all_imports(self) -> List[importsymbol.Import]:
        all_imports = [imported for imported in self.symbols.values()
                       if (isinstance(imported, importsymbol.Import)
                           and not isinstance(imported, importsymbol.BuiltinImport))]
        index = 0
        while index < len(all_imports):
            imported = all_imports[index]
            for inner in imported.all_symbols.values():
                if (isinstance(inner, importsymbol.Import)
                        and not isinstance(inner, importsymbol.BuiltinImport)
                        and inner not in all_imports):
                    all_imports.append(inner)
            index += 1

        return all_imports

    def _check_other_implemented_standards(self):
        other_standards = supportedstandard.neo_standards.copy()
        # verify only standards that were not mentioned
        for standard in self.standards:
            if standard in supportedstandard.neo_standards:
                other_standards.pop(standard)

        # gets a list of all events
        events = [symbol for symbol in self.symbols.copy().values() if isinstance(symbol, Event)]
        for imported in self._get_all_imports():
            events.extend([event for event in imported.all_symbols.values() if isinstance(event, Event)])

        # verify if the methods and events that were implemented corresponds to a standard
        for standard in other_standards:

            # verify the methods
            methods_implemented = True
            standard_methods = other_standards[standard].methods
            index = 0

            while methods_implemented and index < len(standard_methods):
                standard_method = standard_methods[index]
                method_id = standard_method.external_name
                found_methods = self.get_methods_by_display_name(method_id)

                methods_implemented = any(
                    other_standards[standard].match_definition(standard_method, method) for method in found_methods
                )
                index += 1

            if not methods_implemented:
                continue    # if even one of the methods was not implemented, then check the next standard

            # verify the events
            events_implemented = True
            standard_events = other_standards[standard].events
            index = 0

            while events_implemented and index < len(standard_events):
                standard_event = standard_events[index]
                events_implemented = any(
                    (event.name == standard_event.name and
                     other_standards[standard].match_definition(standard_event, event)) for event in events
                )
                index += 1

            if not events_implemented:
                continue    # if even one of the events was not implemented, then check the next standard

            self.standards.append(standard)



import unittest

from boa3.boa3 import Boa3
from boa3.exception import CompilerError
from boa3.model.type.type import Type
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.TestExecutionException import TestExecutionException
from boa3_test.tests.test_classes.testengine import TestEngine


class TestBuiltinMethod(BoaTest):
    default_folder: str = 'test_sc/built_in_methods_test'

    # region len test

    def test_len_of_tuple(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x02'
            + b'\x00'
            + Opcode.PUSH3      # a = (1, 2, 3)
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.LDLOC0     # b = len(a)
            + Opcode.SIZE
            + Opcode.STLOC1
            + Opcode.LDLOC1     # return b
            + Opcode.RET
        )
        path = self.get_contract_path('LenTuple.py')

        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(3, result)

    def test_len_of_list(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x02'
            + b'\x00'
            + Opcode.PUSH3      # a = [1, 2, 3]
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3      # array length
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.LDLOC0     # b = len(a)
            + Opcode.SIZE
            + Opcode.STLOC1
            + Opcode.LDLOC1     # return b
            + Opcode.RET
        )
        path = self.get_contract_path('LenList.py')

        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(3, result)

    def test_len_of_str(self):
        input = 'just a test'
        byte_input = String(input).to_bytes()
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHDATA1            # push the bytes
            + Integer(len(byte_input)).to_byte_array()
            + byte_input
            + Opcode.STLOC0
            + Opcode.PUSHDATA1            # push the bytes
            + Integer(len(byte_input)).to_byte_array()
            + byte_input
            + Opcode.SIZE
            + Opcode.RET
        )
        path = self.get_contract_path('LenString.py')

        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(11, result)

    def test_len_of_bytes(self):
        path = self.get_contract_path('LenBytes.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(3, result)

    def test_len_of_no_collection(self):
        path = self.get_contract_path('LenMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_len_too_many_parameters(self):
        path = self.get_contract_path('LenTooManyParameters.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_len_too_few_parameters(self):
        path = self.get_contract_path('LenTooFewParameters.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    # endregion

    # region append test

    def test_append_tuple(self):
        path = self.get_contract_path('AppendTuple.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_append_sequence(self):
        path = self.get_contract_path('AppendSequence.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_append_mutable_sequence(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH3      # a = [1, 2, 3]
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.LDLOC0     # a.append(4)
            + Opcode.PUSH4
            + Opcode.OVER
            + Opcode.ISTYPE
            + Type.bytearray.stack_item
            + Opcode.JMPIFNOT
            + Integer(5).to_byte_array(min_length=1)
            + Opcode.CAT
            + Opcode.JMP
            + Integer(5).to_byte_array(min_length=1)
            + Opcode.APPEND
            + Opcode.JMP
            + Integer(3).to_byte_array(min_length=1)
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )
        path = self.get_contract_path('AppendMutableSequence.py')

        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'append_example')
        self.assertEqual([1, 2, 3, 4], result)

    def test_append_mutable_sequence_with_builtin(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH3      # a = [1, 2, 3]
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.LDLOC0     # a.append(4)
            + Opcode.PUSH4
            + Opcode.OVER
            + Opcode.ISTYPE
            + Type.bytearray.stack_item
            + Opcode.JMPIFNOT
            + Integer(5).to_byte_array(min_length=1)
            + Opcode.CAT
            + Opcode.JMP
            + Integer(5).to_byte_array(min_length=1)
            + Opcode.APPEND
            + Opcode.JMP
            + Integer(3).to_byte_array(min_length=1)
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )
        path = self.get_contract_path('AppendMutableSequenceBuiltinCall.py')

        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'append_example')
        self.assertEqual([1, 2, 3, 4], result)

    def test_append_too_many_parameters(self):
        path = self.get_contract_path('AppendTooManyParameters.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_append_too_few_parameters(self):
        path = self.get_contract_path('AppendTooFewParameters.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    # endregion

    # region clear test

    def test_clear_tuple(self):
        path = self.get_contract_path('ClearTuple.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_clear_mutable_sequence(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH3      # a = [1, 2, 3]
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.LDLOC0     # a.clear()
            + Opcode.DUP
            + Opcode.ISTYPE
            + Type.bytearray.stack_item
            + Opcode.JMPIFNOT
            + Integer(9).to_byte_array(min_length=1)
            + Opcode.DROP
            + Opcode.PUSHDATA1
            + Integer(0).to_byte_array(min_length=1)
            + Opcode.CONVERT
            + Type.bytearray.stack_item
            + Opcode.JMP
            + Integer(5).to_byte_array(min_length=1)
            + Opcode.CLEARITEMS
            + Opcode.JMP
            + Integer(3).to_byte_array(min_length=1)
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )
        path = self.get_contract_path('ClearMutableSequence.py')

        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'clear_example')
        self.assertEqual([], result)

    def test_clear_mutable_sequence_with_builtin(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH3      # a = [1, 2, 3]
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.LDLOC0     # MutableSequence.clear(a)
            + Opcode.DUP
            + Opcode.ISTYPE
            + Type.bytearray.stack_item
            + Opcode.JMPIFNOT
            + Integer(9).to_byte_array(min_length=1)
            + Opcode.DROP
            + Opcode.PUSHDATA1
            + Integer(0).to_byte_array(min_length=1)
            + Opcode.CONVERT
            + Type.bytearray.stack_item
            + Opcode.JMP
            + Integer(5).to_byte_array(min_length=1)
            + Opcode.CLEARITEMS
            + Opcode.JMP
            + Integer(3).to_byte_array(min_length=1)
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )
        path = self.get_contract_path('ClearMutableSequenceBuiltinCall.py')

        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'clear_example')
        self.assertEqual([], result)

    def test_clear_too_many_parameters(self):
        path = self.get_contract_path('ClearTooManyParameters.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_clear_too_few_parameters(self):
        path = self.get_contract_path('ClearTooFewParameters.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    # endregion

    # region reverse test

    def test_reverse_tuple(self):
        path = self.get_contract_path('ReverseTuple.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_reverse_mutable_sequence(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH3      # a = [1, 2, 3]
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.LDLOC0     # a.reverse()
            + Opcode.REVERSEITEMS
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )
        path = self.get_contract_path('ReverseMutableSequence.py')

        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([3, 2, 1], result)

    def test_reverse_mutable_sequence_with_builtin(self):
        path = self.get_contract_path('ReverseMutableSequenceBuiltinCall.py')
        output = Boa3.compile(path)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main',
                                         expected_result_type=bytes)
        self.assertEqual(b'\x03\x02\x01', result)

    def test_reverse_too_many_parameters(self):
        path = self.get_contract_path('ReverseTooManyParameters.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_reverse_too_few_parameters(self):
        path = self.get_contract_path('ReverseTooFewParameters.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    # endregion

    # region extend test

    def test_extend_tuple(self):
        path = self.get_contract_path('ExtendTuple.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_extend_sequence(self):
        path = self.get_contract_path('ExtendSequence.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_extend_mutable_sequence(self):
        path = self.get_contract_path('ExtendMutableSequence.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([1, 2, 3, 4, 5, 6], result)

    def test_extend_mutable_sequence_with_builtin(self):
        path = self.get_contract_path('ExtendMutableSequenceBuiltinCall.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([1, 2, 3, 4, 5, 6], result)

    def test_extend_too_many_parameters(self):
        path = self.get_contract_path('ExtendTooManyParameters.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_extend_too_few_parameters(self):
        path = self.get_contract_path('ExtendTooFewParameters.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    # endregion

    # region to_script_hash test

    def test_script_hash_int(self):
        from boa3.neo import to_script_hash
        script_hash = to_script_hash(Integer(123).to_byte_array())

        path = self.get_contract_path('ScriptHashInt.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(script_hash, result)

    def test_script_hash_int_with_builtin(self):
        from boa3.neo import to_script_hash
        script_hash = to_script_hash(Integer(123).to_byte_array())

        path = self.get_contract_path('ScriptHashIntBuiltinCall.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(script_hash, result)

    def test_script_hash_str(self):
        from boa3.neo import to_script_hash
        script_hash = to_script_hash(String('123').to_bytes())

        path = self.get_contract_path('ScriptHashStr.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main',
                                         expected_result_type=bytes)
        self.assertEqual(script_hash, result)

    def test_script_hash_str_with_builtin(self):
        from boa3.neo import to_script_hash
        script_hash = to_script_hash(String('123').to_bytes())

        path = self.get_contract_path('ScriptHashStrBuiltinCall.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main',
                                         expected_result_type=bytes)
        self.assertEqual(script_hash, result)

    def test_script_hash_variable(self):
        path = self.get_contract_path('ScriptHashVariable.py')
        engine = TestEngine()
        with self.assertRaises(TestExecutionException):
            self.run_smart_contract(engine, path, 'Main', 123)

        from boa3.neo import to_script_hash
        from base58 import b58encode
        value = b58encode(Integer(123).to_byte_array())
        if isinstance(value, int):
            value = Integer(value).to_byte_array()

        script_hash = to_script_hash(value)

        result = self.run_smart_contract(engine, path, 'Main', value,
                                         expected_result_type=bytes)
        self.assertEqual(script_hash, result)

    def test_script_hash_variable_with_builtin(self):
        path = self.get_contract_path('ScriptHashVariableBuiltinCall.py')
        from boa3.neo import to_script_hash
        from base58 import b58encode
        engine = TestEngine()

        script_hash = to_script_hash('123')
        result = self.run_smart_contract(engine, path, 'Main', '123',
                                         expected_result_type=bytes)
        self.assertEqual(script_hash, result)

        value = b58encode('123')
        if isinstance(value, int):
            value = Integer(value).to_byte_array()

        script_hash = to_script_hash(value)
        result = self.run_smart_contract(engine, path, 'Main', value,
                                         expected_result_type=bytes)
        self.assertEqual(script_hash, result)

    def test_script_hash_too_many_parameters(self):
        path = self.get_contract_path('ScriptHashTooManyParameters.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_script_hash_too_few_parameters(self):
        path = self.get_contract_path('ScriptHashTooFewParameters.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_script_hash_mismatched_types(self):
        path = self.get_contract_path('ScriptHashMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_script_hash_builtin_mismatched_types(self):
        path = self.get_contract_path('ScriptHashBuiltinMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    # endregion

    # region to_bytes test

    def test_int_to_bytes(self):
        path = self.get_contract_path('IntToBytes.py')
        engine = TestEngine()

        value = Integer(123).to_byte_array()
        result = self.run_smart_contract(engine, path, 'int_to_bytes',
                                         expected_result_type=bytes)
        self.assertEqual(value, result)

    def test_int_zero_to_bytes(self):
        path = self.get_contract_path('IntZeroToBytes.py')
        engine = TestEngine()

        value = Integer(0).to_byte_array(min_length=1)
        result = self.run_smart_contract(engine, path, 'int_to_bytes',
                                         expected_result_type=bytes)
        self.assertEqual(value, result)

    def test_int_to_bytes_with_builtin(self):
        path = self.get_contract_path('IntToBytesWithBuiltin.py')
        engine = TestEngine()

        value = Integer(123).to_byte_array()
        result = self.run_smart_contract(engine, path, 'int_to_bytes',
                                         expected_result_type=bytes)
        self.assertEqual(value, result)

    def test_int_to_bytes_as_parameter(self):
        path = self.get_contract_path('IntToBytesAsParameter.py')
        engine = TestEngine()

        self.run_smart_contract(engine, path, 'int_to_bytes', 1111,
                                expected_result_type=bytes)
        from boa3.neo3.vm import VMState
        # return is Void, checking to see if there is no error
        self.assertEqual(engine.vm_state, VMState.HALT)

    def test_str_to_bytes(self):
        value = String('123').to_bytes()
        expected_output = (
            Opcode.PUSHDATA1        # '123'.to_bytes()
            + Integer(len(value)).to_byte_array(min_length=1)
            + value
            + Opcode.RET
        )

        path = self.get_contract_path('StrToBytes.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'str_to_bytes',
                                         expected_result_type=bytes)
        self.assertEqual(value, result)

    def test_str_to_bytes_with_builtin(self):
        value = String('123').to_bytes()
        expected_output = (
            Opcode.PUSHDATA1        # str.to_bytes('123')
            + Integer(len(value)).to_byte_array(min_length=1)
            + value
            + Opcode.RET
        )

        path = self.get_contract_path('StrToBytesWithBuiltin.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'str_to_bytes',
                                         expected_result_type=bytes)
        self.assertEqual(value, result)

    def test_to_bytes_mismatched_types(self):
        path = self.get_contract_path('ToBytesMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    # endregion

    # region print test

    def test_print_int(self):
        path = self.get_contract_path('PrintInt.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'Main')
        self.assertIsVoid(result)

    def test_print_str(self):
        path = self.get_contract_path('PrintStr.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'Main')
        self.assertIsVoid(result)

    def test_print_bytes(self):
        path = self.get_contract_path('PrintBytes.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'Main')
        self.assertIsVoid(result)

    def test_print_bool(self):
        path = self.get_contract_path('PrintBool.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'Main')
        self.assertIsVoid(result)

    def test_print_list(self):
        path = self.get_contract_path('PrintList.py')
        self.assertCompilerLogs(CompilerError.NotSupportedOperation, path)

    def test_print_many_values(self):
        path = self.get_contract_path('PrintManyValues.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'Main')
        self.assertIsVoid(result)

    def test_print_no_args(self):
        path = self.get_contract_path('PrintNoArgs.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'Main')
        self.assertIsVoid(result)

    def test_print_missing_outer_function_return(self):
        path = self.get_contract_path('PrintIntMissingFunctionReturn.py')
        self.assertCompilerLogs(CompilerError.MissingReturnStatement, path)

    # endregion

    # region isinstance test

    def test_isinstance_int_literal(self):
        path = self.get_contract_path('IsInstanceIntLiteral.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(True, result)

    def test_isinstance_int_variable(self):
        path = self.get_contract_path('IsInstanceIntVariable.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 10)
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', False)
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', 'string')
        self.assertEqual(False, result)

    def test_isinstance_bool_literal(self):
        path = self.get_contract_path('IsInstanceBoolLiteral.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(False, result)

    def test_isinstance_bool_variable(self):
        path = self.get_contract_path('IsInstanceBoolVariable.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 10)
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', False)
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', 'string')
        self.assertEqual(False, result)

    def test_isinstance_str_literal(self):
        path = self.get_contract_path('IsInstanceStrLiteral.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(True, result)

    def test_isinstance_str_variable(self):
        path = self.get_contract_path('IsInstanceStrVariable.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 10)
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', False)
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', 'string')
        self.assertEqual(True, result)

    def test_isinstance_list_literal(self):
        path = self.get_contract_path('IsInstanceListLiteral.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(True, result)

    def test_isinstance_tuple_literal(self):
        path = self.get_contract_path('IsInstanceTupleLiteral.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(True, result)

    def test_isinstance_tuple_variable(self):
        path = self.get_contract_path('IsInstanceTupleVariable.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 10)
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', False)
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', 'string')
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', [])
        self.assertEqual(True, result)

    def test_isinstance_many_types(self):
        path = self.get_contract_path('IsInstanceManyTypes.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 10)
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', False)
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', 'string')
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', {})
        self.assertEqual(True, result)

    def test_isinstance_many_types_with_class(self):
        path = self.get_contract_path('IsInstanceManyTypesWithClass.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 42)
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', bytes(10))
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', [])
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', 'some string')
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', bytes(20))
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', None)
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', {1: 2, 2: 4})
        self.assertEqual(True, result)

    def test_isinstance_uint160(self):
        path = self.get_contract_path('IsInstanceUInt160.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', bytes(10))
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', bytes(20))
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', bytes(30))
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', 42)
        self.assertEqual(False, result)

    def test_isinstance_uint256(self):
        path = self.get_contract_path('IsInstanceUInt256.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', bytes(10))
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'main', bytes(20))
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'main', bytes(30))
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'main', bytes(32))
        self.assertEqual(True, result)

    # endregion

    # region exit test

    def test_exit(self):
        path = self.get_contract_path('Exit.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', False)
        self.assertEqual(123, result)

        with self.assertRaisesRegex(TestExecutionException, self.ABORTED_CONTRACT_MSG):
            self.run_smart_contract(engine, path, 'main', True)

    # endregion

    # region max test

    def test_max_int(self):
        path = self.get_contract_path('MaxInt.py')
        engine = TestEngine()

        value1 = 10
        value2 = 999
        expected_result = max(value1, value2)
        result = self.run_smart_contract(engine, path, 'main', value1, value2)
        self.assertEqual(expected_result, result)

    def test_max_int_more_arguments(self):
        path = self.get_contract_path('MaxIntMoreArguments.py')
        engine = TestEngine()

        numbers = 4, 1, 16, 8, 2
        expected_result = max(numbers)
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(expected_result, result)

    def test_max_str(self):
        path = self.get_contract_path('MaxStr.py')
        engine = TestEngine()

        value1 = 'foo'
        value2 = 'bar'
        expected_result = max(value1, value2)
        result = self.run_smart_contract(engine, path, 'main', value1, value2)
        self.assertEqual(expected_result, result)

        result = self.run_smart_contract(engine, path, 'main', value2, value1)
        self.assertEqual(expected_result, result)

        value1 = 'alg'
        value2 = 'al'
        expected_result = max(value1, value2)
        result = self.run_smart_contract(engine, path, 'main', value1, value2)
        self.assertEqual(expected_result, result)

        value = 'some string'
        expected_result = max(value, value)
        result = self.run_smart_contract(engine, path, 'main', value, value)
        self.assertEqual(expected_result, result)

    def test_max_str_more_arguments(self):
        path = self.get_contract_path('MaxStrMoreArguments.py')
        engine = TestEngine()

        values = 'Lorem', 'ipsum', 'dolor', 'sit', 'amet'
        expected_result = max(values)
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(expected_result, result)

    def test_max_bytes(self):
        path = self.get_contract_path('MaxBytes.py')
        engine = TestEngine()

        value1 = b'foo'
        value2 = b'bar'
        expected_result = max(value1, value2)
        result = self.run_smart_contract(engine, path, 'main', value1, value2, expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        result = self.run_smart_contract(engine, path, 'main', value2, value1, expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        value1 = b'alg'
        value2 = b'al'
        expected_result = max(value1, value2)
        result = self.run_smart_contract(engine, path, 'main', value1, value2, expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        value = b'some string'
        expected_result = max(value, value)
        result = self.run_smart_contract(engine, path, 'main', value, value, expected_result_type=bytes)
        self.assertEqual(expected_result, result)

    def test_max_bytes_more_arguments(self):
        path = self.get_contract_path('MaxBytesMoreArguments.py')
        engine = TestEngine()

        values = b'Lorem', b'ipsum', b'dolor', b'sit', b'amet'
        expected_result = max(values)
        result = self.run_smart_contract(engine, path, 'main', expected_result_type=bytes)
        self.assertEqual(expected_result, result)

    def test_max_too_few_parameters(self):
        path = self.get_contract_path('MaxTooFewParameters.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_max_mismatched_types(self):
        path = self.get_contract_path('MaxMismatchedTypes.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    # endregion

    # region min test

    def test_min_int(self):
        path = self.get_contract_path('MinInt.py')
        engine = TestEngine()

        val1 = 50
        val2 = 1
        expected_result = min(val1, val2)
        result = self.run_smart_contract(engine, path, 'main', val1, val2)
        self.assertEqual(expected_result, result)

    def test_min_int_more_arguments(self):
        path = self.get_contract_path('MinIntMoreArguments.py')
        engine = TestEngine()

        numbers = 2, 8, 1, 4, 16
        expected_result = min(numbers)
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(expected_result, result)

    def test_min_str(self):
        path = self.get_contract_path('MinStr.py')
        engine = TestEngine()

        value1 = 'foo'
        value2 = 'bar'
        expected_result = min(value1, value2)
        result = self.run_smart_contract(engine, path, 'main', value1, value2)
        self.assertEqual(expected_result, result)

        result = self.run_smart_contract(engine, path, 'main', value2, value1)
        self.assertEqual(expected_result, result)

        value1 = 'alg'
        value2 = 'al'
        expected_result = min(value1, value2)
        result = self.run_smart_contract(engine, path, 'main', value1, value2)
        self.assertEqual(expected_result, result)

        value = 'some string'
        expected_result = min(value, value)
        result = self.run_smart_contract(engine, path, 'main', value, value)
        self.assertEqual(expected_result, result)

    def test_min_str_more_arguments(self):
        path = self.get_contract_path('MinStrMoreArguments.py')
        engine = TestEngine()

        values = 'Lorem', 'ipsum', 'dolor', 'sit', 'amet'
        expected_result = min(values)
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(expected_result, result)

    def test_min_bytes(self):
        path = self.get_contract_path('MinBytes.py')
        engine = TestEngine()

        value1 = b'foo'
        value2 = b'bar'
        expected_result = min(value1, value2)
        result = self.run_smart_contract(engine, path, 'main', value1, value2, expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        result = self.run_smart_contract(engine, path, 'main', value2, value1, expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        value1 = b'alg'
        value2 = b'al'
        expected_result = min(value1, value2)
        result = self.run_smart_contract(engine, path, 'main', value1, value2, expected_result_type=bytes)
        self.assertEqual(expected_result, result)

        value = b'some string'
        expected_result = min(value, value)
        result = self.run_smart_contract(engine, path, 'main', value, value, expected_result_type=bytes)
        self.assertEqual(expected_result, result)

    def test_min_bytes_more_arguments(self):
        path = self.get_contract_path('MinBytesMoreArguments.py')
        engine = TestEngine()

        values = b'Lorem', b'ipsum', b'dolor', b'sit', b'amet'
        expected_result = min(values)
        result = self.run_smart_contract(engine, path, 'main', expected_result_type=bytes)
        self.assertEqual(expected_result, result)

    def test_min_too_few_parameters(self):
        path = self.get_contract_path('MinTooFewParameters.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_min_mismatched_types(self):
        path = self.get_contract_path('MinMismatchedTypes.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    # endregion

    # region abs test

    def test_abs(self):
        path = self.get_contract_path('Abs.py')
        engine = TestEngine()

        val = 10
        expected_result = abs(val)
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(expected_result, result)

        expected_result = abs(-1)
        result = self.run_smart_contract(engine, path, 'main', -1)
        self.assertEqual(expected_result, result)

        expected_result = abs(1)
        result = self.run_smart_contract(engine, path, 'main', 1)
        self.assertEqual(expected_result, result)

    def test_abs_bytes(self):
        path = self.get_contract_path('AbsMismatchedTypesBytes.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_abs_string(self):
        path = self.get_contract_path('AbsMismatchedTypesString.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_abs_too_few_parameters(self):
        path = self.get_contract_path('AbsTooFewParameters.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_abs_too_many_parameters(self):
        path = self.get_contract_path('AbsTooManyParameters.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    # endregion

    # region sum test

    def test_sum(self):
        path = self.get_contract_path('Sum.py')
        engine = TestEngine()

        val = [1, 2, 3, 4]
        expected_result = sum(val)
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(expected_result, result)

        val = list(range(10, 20, 2))
        expected_result = sum(val)
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(expected_result, result)

    def test_sum_with_start(self):
        path = self.get_contract_path('SumWithStart.py')
        engine = TestEngine()

        val = [1, 2, 3, 4]
        expected_result = sum(val, 10)
        result = self.run_smart_contract(engine, path, 'main', val, 10)
        self.assertEqual(expected_result, result)

        val = list(range(10, 20, 2))
        expected_result = sum(val, 20)
        result = self.run_smart_contract(engine, path, 'main', val, 20)
        self.assertEqual(expected_result, result)

    def test_sum_mismatched_type(self):
        path = self.get_contract_path('SumMismatchedTypes.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_sum_too_few_parameters(self):
        path = self.get_contract_path('SumTooFewParameters.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_sum_too_many_parameters(self):
        path = self.get_contract_path('SumTooManyParameters.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    # endregion

    # region str split test

    def test_str_split(self):
        path = self.get_contract_path('StrSplit.py')
        engine = TestEngine()

        string = '1#2#3#4'
        separator = '#'
        maxsplit = 2
        expected_result = string.split(separator, maxsplit)
        result = self.run_smart_contract(engine, path, 'main', string, separator, maxsplit)
        self.assertEqual(expected_result, result)

        string = '1#2#3#4'
        separator = '#'
        maxsplit = 1
        expected_result = string.split(separator, maxsplit)
        result = self.run_smart_contract(engine, path, 'main', string, separator, maxsplit)
        self.assertEqual(expected_result, result)

        string = '1#2#3#4'
        separator = '#'
        maxsplit = 0
        expected_result = string.split(separator, maxsplit)
        result = self.run_smart_contract(engine, path, 'main', string, separator, maxsplit)
        self.assertEqual(expected_result, result)

        string = 'unit123test123str123split'
        separator = '123'
        maxsplit = 1
        expected_result = string.split(separator, maxsplit)
        result = self.run_smart_contract(engine, path, 'main', string, separator, maxsplit)
        self.assertEqual(expected_result, result)

    def test_str_split_maxsplit_default(self):
        path = self.get_contract_path('StrSplitMaxsplitDefault.py')
        engine = TestEngine()

        string = '1#2#3#4'
        separator = '#'
        expected_result = string.split(separator)
        result = self.run_smart_contract(engine, path, 'main', string, separator)
        self.assertEqual(expected_result, result)

        string = 'unit123test123str123split'
        separator = '123'
        expected_result = string.split(separator)
        result = self.run_smart_contract(engine, path, 'main', string, separator)
        self.assertEqual(expected_result, result)

    def test_str_split_default(self):
        path = self.get_contract_path('StrSplitSeparatorDefault.py')
        engine = TestEngine()

        string = '1 2 3 4'
        expected_result = string.split()
        result = self.run_smart_contract(engine, path, 'main', string)
        self.assertEqual(expected_result, result)

    # endregion

    # region count test

    def test_count_list_int(self):
        path = self.get_contract_path('CountListInt.py')
        engine = TestEngine()

        list_ = [1, 2, 3, 4, 1, 1, 0]
        expected_result = list_.count(1)
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(expected_result, result)

    def test_count_list_str(self):
        path = self.get_contract_path('CountListStr.py')
        engine = TestEngine()

        list_ = ['unit', 'test', 'unit', 'unit', 'random', 'string']
        expected_result = list_.count('unit')
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(expected_result, result)

    def test_count_list_bytes(self):
        path = self.get_contract_path('CountListBytes.py')
        engine = TestEngine()

        list_ = [b'unit', b'test', b'unit', b'unit', b'random', b'string']
        expected_result = list_.count(b'unit')
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(expected_result, result)

    def test_count_list_different_primitive_types(self):
        path = self.get_contract_path('CountListDifferentPrimitiveTypes.py')
        engine = TestEngine()

        list_ = [b'unit', 'test', b'unit', b'unit', 123, 123, True, False]
        expected_result = list_.count(b'unit'), list_.count('test'), list_.count(123)
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(expected_result, tuple(result))

    def test_count_list_different_any_types(self):
        path = self.get_contract_path('CountListAnyType.py')
        engine = TestEngine()

        mixed_list = [[b'unit', b'unit'], [123, 123], [True, False], [True, False], [b'unit', 'test'], 'not list']

        count1 = mixed_list.count([b'unit', 'test'])
        count2 = mixed_list.count([123, 123])
        count3 = mixed_list.count([True, False])
        count4 = mixed_list.count(['random value', 'random value', 'random value'])
        expected_result = [count1, count2, count3, count4]

        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(expected_result, result)

    def test_count_list_only_sequences(self):
        path = self.get_contract_path('CountListOnlySequences.py')
        engine = TestEngine()

        mixed_list = [[b'unit', b'unit'], [123, 123], [True, False], [True, False], [b'unit', 'test']]

        count1 = mixed_list.count([b'unit', 'test'])
        count2 = mixed_list.count([123, 123])
        count3 = mixed_list.count([True, False])
        count4 = mixed_list.count(['random value', 'random value', 'random value'])
        expected_result = [count1, count2, count3, count4]

        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(expected_result, result)

    def test_count_list_empty(self):
        path = self.get_contract_path('CountListEmpty.py')
        engine = TestEngine()

        list_ = []
        expected_result = list_.count(1)
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(expected_result, result)

    def test_count_tuple_int(self):
        path = self.get_contract_path('CountTupleInt.py')
        engine = TestEngine()

        tuple_ = (1, 2, 3, 4, 1, 1, 0)
        expected_result = tuple_.count(1)
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(expected_result, result)

    def test_count_tuple_str(self):
        path = self.get_contract_path('CountTupleStr.py')
        engine = TestEngine()

        tuple_ = ('unit', 'test', 'unit', 'unit', 'random', 'string')
        expected_result = tuple_.count('unit')
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(expected_result, result)

    def test_count_tuple_bytes(self):
        path = self.get_contract_path('CountTupleBytes.py')
        engine = TestEngine()

        tuple_ = (b'unit', b'test', b'unit', b'unit', b'random', b'string')
        expected_result = tuple_.count(b'unit')
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(expected_result, result)

    def test_count_tuple_different_types(self):
        path = self.get_contract_path('CountTupleDifferentTypes.py')
        engine = TestEngine()

        tuple_ = (b'unit', 'test', b'unit', b'unit', 123, 123, True, False)
        expected_result = tuple_.count(b'unit'), tuple_.count('test'), tuple_.count(123)
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(expected_result, tuple(result))

    def test_count_tuple_different_non_primitive_types(self):
        path = self.get_contract_path('CountTupleDifferentNonPrimitiveTypes.py')
        engine = TestEngine()

        mixed_list = ([b'unit', 'test'], [b'unit', b'unit'], [123, 123], [True, False], [True, False])

        count1 = mixed_list.count([b'unit', 'test'])
        count2 = mixed_list.count([123, 123])
        count3 = mixed_list.count([True, False])
        expected_result = [count1, count2, count3]

        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(expected_result, result)

    def test_count_tuple_empty(self):
        path = self.get_contract_path('CountTupleEmpty.py')
        engine = TestEngine()

        tuple_ = ()
        expected_result = tuple_.count(1)
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(expected_result, result)

    def test_count_range(self):
        path = self.get_contract_path('CountRange.py')
        engine = TestEngine()

        expected_result = range(10).count(1)
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(expected_result, result)

    def test_count_sequence_too_many_parameters(self):
        path = self.get_contract_path('CountSequenceTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_count_sequence_too_few_parameters(self):
        path = self.get_contract_path('CountSequenceTooFewArguments.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_count_str(self):
        path = self.get_contract_path('CountStr.py')
        engine = TestEngine()

        str_ = 'eeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee'
        substr = 'e'
        start = 0
        end = 43
        expected_result = str_.count(substr, start, end)
        result = self.run_smart_contract(engine, path, 'main', str_, substr, start, end)
        self.assertEqual(expected_result, result)

        str_ = 'eeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee'
        substr = 'e'
        start = 0
        end = 1000
        expected_result = str_.count(substr, start, end)
        result = self.run_smart_contract(engine, path, 'main', str_, substr, start, end)
        self.assertEqual(expected_result, result)

        str_ = 'eeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee'
        substr = 'e'
        start = 0
        end = -1000
        expected_result = str_.count(substr, start, end)
        result = self.run_smart_contract(engine, path, 'main', str_, substr, start, end)
        self.assertEqual(expected_result, result)

        str_ = 'eeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee'
        substr = 'e'
        start = 0
        end = -4
        expected_result = str_.count(substr, start, end)
        result = self.run_smart_contract(engine, path, 'main', str_, substr, start, end)
        self.assertEqual(expected_result, result)

        str_ = 'eeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee'
        substr = 'e'
        start = 23
        end = 43
        expected_result = str_.count(substr, start, end)
        result = self.run_smart_contract(engine, path, 'main', str_, substr, start, end)
        self.assertEqual(expected_result, result)

        str_ = 'eeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee'
        substr = 'e'
        start = -11
        end = 43
        expected_result = str_.count(substr, start, end)
        result = self.run_smart_contract(engine, path, 'main', str_, substr, start, end)
        self.assertEqual(expected_result, result)

        str_ = 'eeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee'
        substr = 'e'
        start = -1000
        end = 43
        expected_result = str_.count(substr, start, end)
        result = self.run_smart_contract(engine, path, 'main', str_, substr, start, end)
        self.assertEqual(expected_result, result)

        str_ = 'eeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee'
        substr = 'e'
        start = 1000
        end = 43
        expected_result = str_.count(substr, start, end)
        result = self.run_smart_contract(engine, path, 'main', str_, substr, start, end)
        self.assertEqual(expected_result, result)

        str_ = 'a string that will be used in the unit test'
        substr = 'string'
        start = 0
        end = 43
        expected_result = str_.count(substr, start, end)
        result = self.run_smart_contract(engine, path, 'main', str_, substr, start, end)
        self.assertEqual(expected_result, result)

    def test_count_str_end_default(self):
        path = self.get_contract_path('CountStrEndDefault.py')
        engine = TestEngine()

        str_ = 'a string that will be used in the unit test'
        substr = 'string'
        start = 0
        expected_result = str_.count(substr, start)
        result = self.run_smart_contract(engine, path, 'main', str_, substr, start)
        self.assertEqual(expected_result, result)

        str_ = 'a string that will be used in the unit test'
        substr = 'string'
        start = 4
        expected_result = str_.count(substr, start)
        result = self.run_smart_contract(engine, path, 'main', str_, substr, start)
        self.assertEqual(expected_result, result)

        str_ = 'eeeeeeeeeee'
        substr = 'e'
        start = 0
        expected_result = str_.count(substr, start)
        result = self.run_smart_contract(engine, path, 'main', str_, substr, start)
        self.assertEqual(expected_result, result)

        str_ = 'eeeeeeeeeee'
        substr = 'e'
        start = 5
        expected_result = str_.count(substr, start)
        result = self.run_smart_contract(engine, path, 'main', str_, substr, start)
        self.assertEqual(expected_result, result)

    def test_count_str_default(self):
        path = self.get_contract_path('CountStrDefault.py')
        engine = TestEngine()

        str_ = 'a string that will be used in the unit test'
        substr = 'string'
        expected_result = str_.count(substr)
        result = self.run_smart_contract(engine, path, 'main', str_, substr)
        self.assertEqual(expected_result, result)

        str_ = 'eeeeeeeeeee'
        substr = 'e'
        expected_result = str_.count(substr)
        result = self.run_smart_contract(engine, path, 'main', str_, substr)
        self.assertEqual(expected_result, result)

    def test_count_str_too_many_parameters(self):
        path = self.get_contract_path('CountStrTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_count_str_too_few_parameters(self):
        path = self.get_contract_path('CountStrTooFewArguments.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    # endregion

    # region super test

    def test_super_with_args(self):
        # TODO: Change when super with args is implemented
        path = self.get_contract_path('SuperWithArgs.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_super_call_method(self):
        path = self.get_contract_path('SuperCallMethod.py')
        engine = TestEngine()

        super_method_expected_result = -20
        arg = 20
        result = self.run_smart_contract(engine, path, 'example_method', arg)
        self.assertEqual(arg, result)

        arg = 30
        result = self.run_smart_contract(engine, path, 'example_method', arg)
        self.assertEqual(arg, result)

        arg = 40
        result = self.run_smart_contract(engine, path, 'example_method', arg)
        self.assertEqual(super_method_expected_result, result)

    # endregion

    # region int test

    # TODO: Int constructor is not accepting more than one method
    @unittest.skip('Int constructor is not accepting more than one method')
    def test_int_str(self):
        path = self.get_contract_path('IntStr.py')
        self.compile_and_save(path)  # it doesn't compile, it isn't implemented yet
        engine = TestEngine()

        value = '0b101'
        base = 0
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = '0B101'
        base = 0
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = '0b101'
        base = 2
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = '0B101'
        base = 2
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = '0o123'
        base = 0
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = '0O123'
        base = 0
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = '0o123'
        base = 8
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = '0O123'
        base = 8
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = '0x123'
        base = 0
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = '0X123'
        base = 0
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = '0x123'
        base = 16
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = '0X123'
        base = 16
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = '123'
        base = 16
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = '11'
        base = 11
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = 'abcdef'
        base = 16
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = 'abcdefg'
        base = 16
        with self.assertRaises(TestExecutionException):
            self.run_smart_contract(engine, path, 'main', value, base)
        with self.assertRaises(ValueError):
            int(value, base)

        value = '0x123'
        base = 8
        with self.assertRaises(TestExecutionException):
            self.run_smart_contract(engine, path, 'main', value, base)
        with self.assertRaises(ValueError):
            int(value, base)

    # TODO: Int constructor is not accepting more than one method
    @unittest.skip('Int constructor is not accepting more than one method')
    def test_int_bytes(self):
        path = self.get_contract_path('IntBytes.py')
        self.compile_and_save(path)  # it doesn't compile, it isn't implemented yet
        engine = TestEngine()

        value = b'0b101'
        base = 0
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = b'0B101'
        base = 0
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = b'0b101'
        base = 2
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = b'0B101'
        base = 2
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = b'0o123'
        base = 0
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = b'0O123'
        base = 0
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = b'0o123'
        base = 8
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = b'0O123'
        base = 8
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = b'0x123'
        base = 0
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = b'0X123'
        base = 0
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = b'0x123'
        base = 16
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = b'0X123'
        base = 16
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = b'123'
        base = 16
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = b'11'
        base = 11
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = b'abcdef'
        base = 16
        result = self.run_smart_contract(engine, path, 'main', value, base)
        self.assertEqual(int(value, base), result)

        value = b'abcdefg'
        base = 16
        with self.assertRaises(TestExecutionException):
            self.run_smart_contract(engine, path, 'main', value, base)
        with self.assertRaises(ValueError):
            int(value, base)

        value = b'0x123'
        base = 8
        with self.assertRaises(TestExecutionException):
            self.run_smart_contract(engine, path, 'main', value, base)
        with self.assertRaises(ValueError):
            int(value, base)

    def test_int_int(self):
        path = self.get_contract_path('IntInt.py')
        engine = TestEngine()

        value = 10
        result = self.run_smart_contract(engine, path, 'main', value)
        self.assertEqual(int(value), result)

        value = -10
        result = self.run_smart_contract(engine, path, 'main', value)
        self.assertEqual(int(value), result)

    def test_int_int_too_many_parameters(self):
        path = self.get_contract_path('IntIntTooManyParameters.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_int_no_parameters(self):
        path = self.get_contract_path('IntNoParameters.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(int(), result)

    # endregion

    # region bool test

    def test_bool(self):
        path = self.get_contract_path('Bool.py')
        engine = TestEngine()

        val = 1
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(bool(val), result)

        val = 'test'
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(bool(val), result)

        val = b'test'
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(bool(val), result)

        val = [1, 2, 3]
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(bool(val), result)

        val = {'1': 2}
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(bool(val), result)

    def test_bool_bytes(self):
        path = self.get_contract_path('BoolBytes.py')
        engine = TestEngine()

        val = b'123'
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(bool(val), result)

        val = b''
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(bool(val), result)

    def test_bool_class(self):
        path = self.get_contract_path('BoolClass.py')
        engine = TestEngine()

        # dunder methods are not supported in neo3-boa
        class Example:
            def __init__(self):
                self.test = 123
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(bool(Example()), result)

    def test_bool_dict(self):
        path = self.get_contract_path('BoolDict.py')
        engine = TestEngine()

        val = {}
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(bool(val), result)

        val = {'a': 123, 'b': 56, 'c': 1}
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(bool(val), result)

    def test_bool_int(self):
        path = self.get_contract_path('BoolInt.py')
        engine = TestEngine()

        val = -1
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(bool(val), result)

        val = 0
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(bool(val), result)

        val = 1
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(bool(val), result)

    def test_bool_list(self):
        path = self.get_contract_path('BoolList.py')
        engine = TestEngine()

        val = []
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(bool(val), result)

        val = [1, 2, 3, 4, 5]
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(bool(val), result)

    def test_bool_range(self):
        path = self.get_contract_path('BoolRange.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'range_0')
        self.assertEqual(bool(range(0)), result)

        result = self.run_smart_contract(engine, path, 'range_not_0')
        self.assertEqual(bool(range(10)), result)

    def test_bool_str(self):
        path = self.get_contract_path('BoolStr.py')
        engine = TestEngine()

        val = 'unit test'
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(bool(val), result)

        val = ''
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(bool(val), result)

    # endregion

    # region list test

    def test_list_any(self):
        path = self.get_contract_path('ListAny.py')
        engine = TestEngine()

        val = [1, 2, 3, 4]
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(list(val), result)

        val = {'a': 1, 'b': '2', 'c': True, 'd': b'01', '12e12e12e12': 123}
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(list(val), result)

        val = 'unit test'
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(list(bytes(val, 'utf-8')), result)

        val = b'unit test'
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(list(val), result)

        val = 123
        with self.assertRaises(TestExecutionException):
            self.run_smart_contract(engine, path, 'main', val)

    def test_list_default(self):
        path = self.get_contract_path('ListDefault.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(list(), result)

    def test_list_sequence(self):
        path = self.get_contract_path('ListSequence.py')
        engine = TestEngine()

        val = [1, 2, 3, 4]
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(list(val), result)

        val = (1, 2, 3, 4)
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(list(val), result)

        val = range(0, 10)
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(list(val), result)

        val = [1, 2, 3, 4]
        result = self.run_smart_contract(engine, path, 'verify_list_unchanged', val)
        new_list = list(val)
        val[0] = val[1]
        self.assertEqual(new_list, result)

    def test_list_sequence_mismatched_return_type(self):
        path = self.get_contract_path('ListSequenceMismatchedReturnType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_list_mapping(self):
        path = self.get_contract_path('ListMapping.py')
        engine = TestEngine()

        val = {'a': 1, 'b': 2, 'c': 3, 'd': 0, '12e12e12e12': 123}
        result = self.run_smart_contract(engine, path, 'main', val)
        # result should be a list of all the keys used in the dictionary, in insertion order
        self.assertEqual(list(val), result)

        val = {'a': 1, 'b': '2', 'c': True, 'd': b'01', '12e12e12e12': 123}
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(list(val), result)

        val = {1: 0, 23: 12, -10: 412, 25: '123'}
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(list(val), result)

        val = {b'123': 123, b'test': 'test', b'unit': 'unit'}
        result = self.run_smart_contract(engine, path, 'main', val)
        for index, item in enumerate(result):
            if isinstance(item, str):
                result[index] = String(item).to_bytes()
        self.assertEqual(list(val), result)

        val = {'a': 1, 'b': '2', 'c': True, 1: 0, 23: 12, 25: 'value'}
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(list(val), result)

    def test_list_mapping_mismatched_return_type(self):
        path = self.get_contract_path('ListMappingMismatchedReturnType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_list_bytes(self):
        path = self.get_contract_path('ListBytes.py')
        engine = TestEngine()

        val = b'unit test'
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(list(val), result)

    def test_list_bytes_mismatched_return_type(self):
        path = self.get_contract_path('ListBytesMismatchedReturnType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_list_str(self):
        path = self.get_contract_path('ListString.py')
        engine = TestEngine()

        val = 'unit test'
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(list(val), result)

    def test_list_str_mismatched_return_type(self):
        path = self.get_contract_path('ListStringMismatchedReturnType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_list_bytes_str(self):
        path = self.get_contract_path('ListBytesString.py')
        engine = TestEngine()

        # If the compiler doesn't have a type hint to know if it's a bytes or str value it will consider it as bytes
        val = 'unit test'
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(list(bytes(val, 'utf-8')), result)

        val = b'unit test'
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(list(val), result)

    # endregion

    # region str test

    def test_str_bytes_str(self):
        path = self.get_contract_path('StrByteString.py')
        engine = TestEngine()

        value = 'test'
        result = self.run_smart_contract(engine, path, 'str_parameter', value)
        self.assertEqual(str(value), result)

        value = b'test'
        result = self.run_smart_contract(engine, path, 'bytes_parameter', value)
        # since bytes and string is the same thing internally it will return 'test' instead of the "b'test'"
        self.assertEqual('test', result)

        result = self.run_smart_contract(engine, path, 'empty_parameter')
        self.assertEqual(str(), result)

    def test_str_int(self):
        path = self.get_contract_path('StrInt.py')
        engine = TestEngine()

        value = 1234567890
        result = self.run_smart_contract(engine, path, 'main', value)
        self.assertEqual(str(value), result)

        value = -1234567890
        result = self.run_smart_contract(engine, path, 'main', value)
        self.assertEqual(str(value), result)

    def test_str_bool(self):
        path = self.get_contract_path('StrBool.py')
        engine = TestEngine()

        value = True
        result = self.run_smart_contract(engine, path, 'main', value)
        self.assertEqual(str(value), result)

        value = False
        result = self.run_smart_contract(engine, path, 'main', value)
        self.assertEqual(str(value), result)

    def test_str_too_many_parameters(self):
        path = self.get_contract_path('StrTooManyParameters.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    # endregion



import json
import os.path
import subprocess
from typing import Any, Dict, List, Optional, Tuple, Union

from boa3 import constants
from boa3.neo import utils as neo_utils
from boa3.neo.smart_contract.VoidType import VoidType
from boa3.neo.vm.type.String import String
from boa3.neo3.core.types import UInt160
from boa3.neo3.vm import VMState, vmstate
from test_runner import neoxp_utils
from test_runner.blockchain import *
from test_runner.blockchain.contractcollection import ContractCollection
from test_runner.blockchain.storagecollection import StorageCollection
from test_runner.neoxp.batch import NeoExpressBatch
from test_runner.neoxp.neoinvoke import NeoInvoke


class NeoTestRunner:
    _FOLDER_NAME = 'test-runner'
    _ROOT_FOLDER = os.path.abspath(os.path.curdir) + os.path.sep + _FOLDER_NAME
    _INVOKE_FILE = f'{_FOLDER_NAME}.neo-invoke.json'
    _BATCH_FILE = f'{_FOLDER_NAME}.batch'
    _CHECKPOINT_FILE = f'{_FOLDER_NAME}.neoxp-checkpoint'
    _NEO_EXPRESS_FILE = f'{_FOLDER_NAME}.neo-express'

    def __init__(self, root_test_folder: str = None):
        self._root = root_test_folder

        self._vm_state: VMState = VMState.NONE
        self._gas_consumed: int = 0
        self._result_stack: List[Any] = []
        self._error_message: Optional[str] = None

        self._notifications: List[Notification] = []
        self._logs: List[Log] = []
        self._contracts: ContractCollection = ContractCollection()
        self._storages: StorageCollection = StorageCollection()

        self._batch = NeoExpressBatch()
        self._init_neo_express()

    @property
    def error(self) -> Optional[str]:
        return self._error_message

    @property
    def vm_state(self) -> VMState:
        return self._vm_state

    @property
    def gas_consumed(self) -> int:
        return self._gas_consumed

    @property
    def result_stack(self) -> List[Any]:
        return self._result_stack.copy()

    @property
    def notifications(self) -> List[Notification]:
        return self._notifications.copy()

    @property
    def logs(self) -> List[Log]:
        return self._logs.copy()

    @property
    def contracts(self) -> ContractCollection:
        return self._contracts

    @property
    def storages(self) -> StorageCollection:
        return self._storages

    def silent_run(self, contract_id: Union[str, bytes], method: str, *arguments: Any):
        contract_hash = self._get_contract_hash_or_name(contract_id)
        invoke = self._get_invoke(contract_hash, method, *arguments)
        self._deploy_contract(contract_id, True)
        self._batch.run_contract(invoke)

    def run(self, contract_id: Union[str, bytes], method: str, *arguments: Any,
            get_storage: bool = False,
            reset_runner: bool = False) -> Any:

        contract_hash = self._get_contract_hash_or_name(contract_id)
        invoke = self._get_invoke(contract_hash, method, *arguments)

        self._batch.mint_block(1)
        self._generate_files(contract_id, invoke)
        self._batch.pop()  # remove the forced block minting from batch for optimization

        cli_args = ['neo-test-runner', self._INVOKE_FILE,
                    '--checkpoint', self._CHECKPOINT_FILE,
                    '--express', self._NEO_EXPRESS_FILE
                    ]

        if get_storage and contract_id in self._contracts:
            calling_contract = self._contracts[contract_id]
            if hasattr(calling_contract, 'name'):
                cli_args.extend(['--storages', calling_contract.name])
        stdout, stderr = self._run_command_line(cli_args)

        if reset_runner:
            self.reset_runner()
        else:
            self.reset_state()

        try:
            result = json.loads(stdout)

            self._update_runner(result)
            if self._error_message is None:
                self._batch.run_contract(invoke)
        except BaseException:
            self._error_message = stdout

        if self._error_message is not None:
            return self._error_message
        else:
            # TODO: convert the result to the return type of the function in the manifest
            return self._result_stack[-1] if len(self._result_stack) > 0 else VoidType

    @property
    def _root(self) -> str:
        return self._ROOT_FOLDER

    @_root.setter
    def _root(self, value: str):
        if isinstance(value, str) and os.path.exists(value) and os.path.isdir(value):
            sep = os.path.sep
            root = os.path.abspath(value) + sep + self._FOLDER_NAME
            self._ROOT_FOLDER = root
            self._INVOKE_FILE = root + sep + NeoTestRunner._INVOKE_FILE
            self._BATCH_FILE = root + sep + NeoTestRunner._BATCH_FILE
            self._CHECKPOINT_FILE = root + sep + NeoTestRunner._CHECKPOINT_FILE
            self._NEO_EXPRESS_FILE = root + sep + NeoTestRunner._NEO_EXPRESS_FILE

        elif not self._INVOKE_FILE.startswith(self._ROOT_FOLDER):
            root_folder = self._ROOT_FOLDER + os.path.sep
            self._INVOKE_FILE = root_folder + NeoTestRunner._INVOKE_FILE
            self._BATCH_FILE = root_folder + NeoTestRunner._BATCH_FILE
            self._CHECKPOINT_FILE = root_folder + NeoTestRunner._CHECKPOINT_FILE
            self._NEO_EXPRESS_FILE = root_folder + NeoTestRunner._NEO_EXPRESS_FILE

    def _get_contract_hash_or_name(self, contract_id: Union[str, bytes]) -> Union[str, UInt160]:
        if isinstance(contract_id, str):
            if contract_id in self._contracts:
                contract = self._contracts[contract_id]
                return contract.name if hasattr(contract, 'name') else contract.script_hash

            try:
                return UInt160.from_string(contract_id)
            except BaseException as e:
                if not contract_id.endswith('.nef'):
                    raise e
                manifest_file = contract_id.replace('.nef', '.manifest.json')
                if os.path.exists(manifest_file):
                    with open(manifest_file) as manifest_output:
                        import json
                        manifest = json.loads(manifest_output.read())
                        return manifest['name']

                raise e
        else:
            return UInt160(contract_id)

    def _generate_root_folder(self):
        if not os.path.exists(self._root):
            os.mkdir(self._root)

    def _generate_files(self, contract_id: Union[str, bytes], invoke: NeoInvoke):
        self._generate_root_folder()

        self._generate_checkpoint(contract_id)
        self._generate_batch_file()
        self._batch.pop()  # remove checkpoint operation from batch after generating the file
        if os.path.exists(contract_id) and contract_id.endswith('.nef') and contract_id not in self.contracts:
            contract = neoxp_utils.get_last_deployed_contract(self._NEO_EXPRESS_FILE)
            contract.path = contract_id
            self._contracts.append(contract)

        self._generate_invoke_file(invoke)

    def _init_neo_express(self):
        self._generate_root_folder()
        if not os.path.exists(self._NEO_EXPRESS_FILE):
            neoxp_utils.create_neo_express_instance(self._NEO_EXPRESS_FILE)
        else:
            neoxp_utils.reset_neo_express_instance(self._NEO_EXPRESS_FILE)

    def _deploy_contract(self, contract_id: str, deploy_and_execute: bool = False):
        if os.path.exists(contract_id) and contract_id.endswith('.nef') and contract_id not in self.contracts:
            self._batch.deploy_contract(contract_id)
            if deploy_and_execute:
                self._generate_batch_file()
                contract = neoxp_utils.get_last_deployed_contract(self._NEO_EXPRESS_FILE)
                contract.path = contract_id
                self._contracts.append(contract)

    def _generate_checkpoint(self, contract_id: Union[str, UInt160]):
        if os.path.exists(contract_id) and contract_id.endswith('.nef'):
            self._deploy_contract(contract_id)
            self._batch.create_neo_express_checkpoint(self._CHECKPOINT_FILE, True)

    def _generate_batch_file(self):
        assert self._batch.write(self._BATCH_FILE)
        neoxp_utils.run_batch(self._NEO_EXPRESS_FILE, self._BATCH_FILE)

    def _generate_invoke_file(self, invoke: NeoInvoke):
        test_engine_args = invoke.to_json()
        param_json = json.dumps(test_engine_args, separators=(',', ':'))
        with open(self._INVOKE_FILE, 'wb+') as json_file:
            json_file.write(String(param_json).to_bytes())

    def _get_invoke(self, contract_hash: UInt160, method: str, *args: Any) -> NeoInvoke:
        return NeoInvoke(contract_hash, method, *args)

    def _update_runner(self, result: Dict[str, Any]):
        self._error_message = result['exception'] if 'exception' in result else None

        if 'state' in result:
            self._vm_state = vmstate.get_vm_state(result['state'])

        if 'gasconsumed' in result:
            self._gas_consumed = int(float(result['gasconsumed']) * 10 ** constants.GAS_DECIMALS)

        if 'stack' in result:
            result_stack = result['stack']
            if isinstance(result_stack, list):
                self._result_stack = [neo_utils.stack_item_from_json(value) for value in result_stack]
            else:
                self._result_stack = [neo_utils.stack_item_from_json(result_stack)]

        if 'notifications' in result:
            json_notifications = result['notifications']
            if not isinstance(json_notifications, list):
                json_notifications = [json_notifications]

            notifications = []
            for n in json_notifications:
                new = Notification.from_json(n)
                if new is not None:
                    notifications.append(new)
            self._notifications = notifications

        if 'logs' in result:
            json_logs = result['logs']
            if not isinstance(json_logs, list):
                json_logs = [json_logs]

            logs = []
            for l in json_logs:
                new = Log.from_json(l)
                if new is not None:
                    logs.append(new)
            self._logs = logs

        if 'storages' in result:
            json_storages = result['storages']
            if not isinstance(json_storages, list):
                json_storages = [json_storages]

            storages = []
            for s in json_storages:
                new = Storage.from_json(s, self._contracts)
                if new is not None:
                    storages.append(new)
            self._storages = storages

    def reset_state(self):
        self._vm_state = VMState.NONE
        self._gas_consumed = 0
        self._result_stack = []
        self._error_message = None

    def reset_runner(self):
        self.reset_state()
        self._notifications.clear()
        self._logs.clear()
        self._contracts.clear()
        self._storages.clear()
        self._batch.clear()
        neoxp_utils.reset_neo_express_instance(self._NEO_EXPRESS_FILE)

    def _run_command_line(self, args: List[str]) -> Tuple[str, str]:
        process = subprocess.Popen(args,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.STDOUT,
                                   text=True)
        return process.communicate()



from boa3.builtin.compile_time import public

from boa3_test.test_sc.import_test.package_with_import import Module


@public
def call_imported_method() -> list:
    a: Module.List[Module.Any] = Module.EmptyList()
    return a


@public
def call_imported_variable() -> int:
    return Module.default_number



from typing import Any, Iterable, List, Set

from boa3.model.type.annotation.uniontype import UnionType
from boa3.model.type.itype import IType
from boa3.model.type.primitive.nonetype import noneType


class OptionalType(UnionType):
    """
    An class used to represent Python Optional annotation type
    """

    def __init__(self, optional_types: Set[IType] = None):
        if optional_types is None:
            union_types = None
        else:
            union_types = optional_types.copy()
            if noneType in optional_types:
                optional_types.remove(noneType)
            else:
                union_types.add(noneType)

        super().__init__(union_types)
        self._identifier = 'Optional'
        self._optional_type = optional_types

    @property
    def identifier(self) -> str:
        return '{0}[{1}]'.format(self._identifier,
                                 ', '.join([t.identifier for t in self._optional_type]))

    @property
    def optional_types(self) -> List[IType]:
        return list(self._optional_type)

    def _is_type_of(self, value: Any) -> bool:
        if not isinstance(self._optional_type, Iterable):
            return False
        if isinstance(value, UnionType):
            return all(self._is_type_of(x) for x in value._union_types)
        return any(t.is_type_of(value) for t in self._union_types)

    @classmethod
    def build(cls, value: Any) -> IType:
        if isinstance(value, Iterable):
            value = set(value)
        else:
            value = {value}
        value.add(noneType)
        return super().build(value)



from typing import Optional

from boa3.model.builtin.classmethod.countsequencemethod import CountSequenceMethod
from boa3.model.type.collection.sequence.sequencetype import SequenceType
from boa3.model.type.itype import IType


class CountSequencePrimitiveMethod(CountSequenceMethod):

    def __init__(self, sequence_type: Optional[SequenceType] = None, arg_value: Optional[IType] = None):
        if not isinstance(sequence_type, SequenceType):
            from boa3.model.type.type import Type
            sequence_type = Type.sequence.build_collection([Type.str, Type.int, Type.bytes])

        super().__init__(sequence_type, arg_value)



from typing import Optional

from boa3.model.operation.operator import Operator
from boa3.model.operation.unary import *
from boa3.model.operation.unary.unaryoperation import UnaryOperation
from boa3.model.type.itype import IType


class UnaryOp:
    # Arithmetic operations
    Positive = Positive()
    Negative = Negative()

    # Logical operations
    Not = BooleanNot()
    BitNot = LogicNot()

    @classmethod
    def validate_type(cls, operator: Operator, operand: IType) -> Optional[UnaryOperation]:
        """
        Gets a unary operation given the operator and the operand type.

        :param operator: unary operator
        :param operand: type of the operand
        :return: The operation if exists. None otherwise;
        :rtype: UnaryOperation or None
        """
        for id, op in vars(cls).items():
            if isinstance(op, UnaryOperation) and op.is_valid(operator, operand):
                return op.build(operand)

    @classmethod
    def get_operation_by_operator(cls, operator: Operator) -> Optional[UnaryOperation]:
        """
        Gets a unary operation given the operator.

        :param operator: unary operator
        :return: The operation if exists. If exists more than one operation with the same operator, returns the first
        found. None otherwise.
        :rtype: UnaryOperation or None
        """
        for id, op in vars(cls).items():
            if isinstance(op, UnaryOperation) and op.operator is operator:
                return op

    @classmethod
    def get_operation(cls, operation: UnaryOperation) -> Optional[UnaryOperation]:
        """
        Gets an unary operation given another operation.

        :param operation: unary operation
        :return: The operation if exists. None otherwise;
        :rtype: UnaryOperation or None
        """
        for id, op in vars(cls).items():
            if type(operation) == type(op):
                return op



from __future__ import annotations

from typing import Dict, List, Optional, Tuple, Union

from boa3 import constants
from boa3.model.builtin.builtin import Builtin, BoaPackage
from boa3.model.builtin.interop.interop import Interop
from boa3.model.builtin.native.nativecontract import NativeContract
from boa3.model.event import Event
from boa3.model.identifiedsymbol import IdentifiedSymbol
from boa3.model.imports.package import Package
from boa3.model.symbol import ISymbol
from boa3.model.type.math import Math
from boa3.model.type.typeutils import TypeUtils

__all__ = ['get_package',
           'get_internal_symbol',
           'CompilerBuiltin'
           ]


def get_package(package_full_path: str) -> Optional[Package]:
    return CompilerBuiltin.instance().get_package(package_full_path)


def get_internal_symbol(symbol_id: str) -> Optional[ISymbol]:
    return CompilerBuiltin.instance().get_internal_symbol(symbol_id)


class CompilerBuiltin:
    _instance = None

    @classmethod
    def instance(cls) -> CompilerBuiltin:
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def __init__(self):
        self.packages: List[Package] = []
        self._events: List[Event] = []

        self._generate_builtin_package('typing', TypeUtils.get_types_from_typing_lib())
        self._generate_builtin_package('math', Math.get_methods_from_math_lib())
        self._generate_builtin_package('boa3.builtin', Builtin.boa_builtins)
        self._set_events(Builtin.builtin_events())
        self._generate_builtin_package('boa3.builtin.contract', Builtin.package_symbols(BoaPackage.Contract))
        self._generate_builtin_package('boa3.builtin.compile_time', Builtin.package_symbols(BoaPackage.CompileTime))
        self._generate_builtin_package('boa3.builtin.interop', Interop.package_symbols)
        self._set_events(Interop.interop_events())
        self._generate_builtin_package('boa3.builtin.nativecontract', NativeContract.package_symbols)
        self._generate_builtin_package('boa3.builtin.type', Builtin.package_symbols(BoaPackage.Type))
        self._generate_builtin_package('boa3.builtin.vm', Builtin.package_symbols(BoaPackage.VM))

    @classmethod
    def reset(cls):
        if cls._instance is not None:
            for event in cls._instance._events:
                event.reset_calls()

    def _set_events(self, events: List[Event]):
        self._events.extend(events)

    def _generate_builtin_package(self, package_full_path: str,
                                  symbols: Union[Dict[str, ISymbol], List[IdentifiedSymbol]] = None):
        if isinstance(symbols, list):
            symbols = {symbol.identifier: symbol for symbol in symbols}
        if not isinstance(symbols, dict):
            symbols = {}

        package_ids = package_full_path.split(constants.ATTRIBUTE_NAME_SEPARATOR)
        cur_package: Package = None

        for package_id in package_ids:
            package = None

            if isinstance(cur_package, Package):
                if package_id in cur_package.inner_packages:
                    package = cur_package.inner_packages[package_id]
            else:
                package = next((root_package for root_package in self.packages
                                if root_package.identifier == package_id),
                               None)

            if not isinstance(package, Package):
                package = Package(identifier=package_id)
                if isinstance(cur_package, Package):
                    cur_package.include_symbol(package_id, package)
                else:
                    self.packages.append(package)

            cur_package = package

        if isinstance(cur_package, Package):
            for symbol_id, symbol in symbols.items():
                cur_package.include_symbol(symbol_id, symbol)

    def get_package(self, package_full_path: str) -> Optional[Package]:
        package_ids = package_full_path.split(constants.ATTRIBUTE_NAME_SEPARATOR)

        cur_package: Package = next((root_package for root_package in self._instance.packages
                                     if root_package.identifier == package_ids[0]),
                                    None)
        if cur_package is None:
            return None

        for package_id in package_ids[1:]:
            if package_id not in cur_package.inner_packages:
                return None

            cur_package = cur_package.inner_packages[package_id]

        return cur_package

    def get_internal_symbol(self, symbol_id: str) -> Optional[ISymbol]:
        packages_stack: List[Tuple[list, int]] = []
        current_list = self._instance.packages
        current_index = 0

        while len(current_list) > current_index or len(packages_stack) > 0:
            if len(current_list) <= current_index:
                # if didn't find in the current list, go back to the previous list search
                # if the stack is empty, it doesn't continue the loop because the while condition
                current_list, current_index = packages_stack.pop()
                if len(current_list) <= current_index:
                    # if the previous list has no elements unchecked, just continue the loop
                    continue

            package = current_list[current_index]
            if symbol_id in package.symbols and not isinstance(package.symbols[symbol_id], Package):
                return package.symbols[symbol_id]
            current_index += 1

            # if the package has internal packages, searches the symbol in these packages before continuing
            internal_packages = [symbol for symbol in package.inner_packages.values()]
            if len(internal_packages) > 0:
                # save the current list and index in the package
                # made this way to avoid recursive function call
                packages_stack.append((current_list, current_index))
                current_list = internal_packages
                current_index = 0

        if symbol_id in TypeUtils.symbols_for_internal_validation:
            return TypeUtils.symbols_for_internal_validation[symbol_id]
        return None



from __future__ import annotations

from typing import Any, List, Optional, Tuple

from boa3.model.builtin.interop.interopinterfacetype import InteropInterfaceType
from boa3.model.method import Method
from boa3.model.type.collection.icollection import ICollectionType
from boa3.model.type.itype import IType
from boa3.neo.vm.opcode.Opcode import Opcode


class IteratorType(InteropInterfaceType, ICollectionType):
    """
    A class used to represent Neo Iterator class
    """

    def __init__(self, collection: ICollectionType = None):
        super().__init__('Iterator')

        if collection is None:
            from boa3.model.type.type import Type
            collection = Type.collection

        self._constructor: Method = None
        self._origin_collection: ICollectionType = collection

        self._methods = None
        self._properties = None

        self.key_type = self._origin_collection.valid_key
        self.item_type = self._origin_collection.item_type

    @property
    def identifier(self) -> str:
        return '{0}[{1}, {2}]'.format(self._identifier, self.valid_key.identifier, self.item_type.identifier)

    def constructor_method(self) -> Optional[Method]:
        return self._constructor

    @property
    def value_type(self) -> IType:
        return self._origin_collection.item_type

    @property
    def valid_key(self) -> IType:
        return self._origin_collection.valid_key

    def is_valid_key(self, key_type: IType) -> bool:
        return self._origin_collection.is_valid_key(key_type)

    @classmethod
    def build(cls, value: Any = None) -> IteratorType:
        if isinstance(value, ICollectionType):
            return IteratorType(value)
        else:
            return _Iterator

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, IteratorType)

    def _init_class_symbols(self):
        super()._init_class_symbols()

        from boa3.model.builtin.interop.iterator.iteratornextmethod import IteratorNextMethod
        from boa3.model.builtin.interop.iterator.getiteratorvalue import IteratorValueProperty

        self._instance_methods['next'] = IteratorNextMethod()

        self._properties['value'] = IteratorValueProperty(self)

    def is_instance_opcodes(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.model.type.classes.pythonclass import PythonClass
        return super(PythonClass, self).is_instance_opcodes()

    def __eq__(self, other) -> bool:
        if not isinstance(other, IteratorType):
            return False
        return self.valid_key == other.valid_key and self.value_type == other.value_type

    def __hash__(self) -> int:
        return self._origin_collection.__hash__()


_Iterator = IteratorType()
